# Copyright (c) 2017 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for the password storage."""

import collections
import io
import json
import os
import stat
from unittest import mock

from cinder.tests.unit import test
from cinder.volume.drivers.datacore import passwd


class FakeFileStorage(object):
    """Mock FileStorage class."""
    def __init__(self):
        self._storage = {
            'resource1': {
                'user1': 'resource1-user1',
                'user2': 'resource1-user2',
            },
            'resource2': {
                'user1': 'resource2-user1',
            }
        }

    def open(self):
        return self

    def load(self):
        return self._storage

    def save(self, storage):
        self._storage = storage

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class PasswordFileStorageTestCase(test.TestCase):
    """Tests for the password storage."""

    def test_get_password(self):
        fake_file_storage = FakeFileStorage()
        passwords = fake_file_storage.load()
        resource = next(iter(passwords.keys()))
        user, expected = next(iter(passwords[resource].items()))

        self._mock_file_storage(fake_file_storage)
        password_storage = passwd.PasswordFileStorage('fake_file_path')

        result = password_storage.get_password(resource, user)
        self.assertEqual(expected, result)

        result = password_storage.get_password(resource.upper(), user)
        self.assertIsNone(result)

    def test_set_password(self):
        fake_file_storage = FakeFileStorage()
        user = 'user3'
        resource1 = 'resource2'
        password1 = 'resource2-user3'
        resource2 = 'resource3'
        password2 = 'resource3-user3'

        self._mock_file_storage(fake_file_storage)
        password_storage = passwd.PasswordFileStorage('fake_file_path')

        password_storage.set_password(resource1, user, password1)
        passwords = fake_file_storage.load()
        self.assertIn(resource1, passwords)
        self.assertIn(user, passwords[resource1])
        self.assertEqual(password1, passwords[resource1][user])

        password_storage.set_password(resource2, user, password2)
        passwords = fake_file_storage.load()
        self.assertIn(resource2, passwords)
        self.assertIn(user, passwords[resource2])
        self.assertEqual(password2, passwords[resource2][user])

    def test_delete_password(self):
        fake_file_storage = FakeFileStorage()
        passwords = fake_file_storage.load()
        resource1, resource2 = 'resource1', 'resource2'
        user1, res1 = next(iter(passwords[resource1].items()))
        user2, res2 = next(iter(passwords[resource2].items()))

        self._mock_file_storage(fake_file_storage)
        password_storage = passwd.PasswordFileStorage('fake_file_path')

        password_storage.delete_password(resource1, user1)
        passwords = fake_file_storage.load()
        self.assertIn(resource1, passwords)
        self.assertNotIn(user1, passwords[resource1])

        password_storage.delete_password(resource2, user2)
        passwords = fake_file_storage.load()
        self.assertNotIn(resource2, passwords)

    def _mock_file_storage(self, fake_file_storage):
        self.mock_object(passwd, 'FileStorage', return_value=fake_file_storage)


class FileStorageTestCase(test.TestCase):
    """Test for the file storage."""

    def test_open(self):
        fake_file_path = 'file_storage.data'
        self.mock_object(passwd.os.path, 'isfile', return_value=True)
        self.mock_object(passwd.os.path, 'isdir', return_value=True)
        mock_open = self.mock_object(passwd, 'open', mock.mock_open())

        file_storage = passwd.FileStorage(fake_file_path)
        file_storage.open()
        mock_open.assert_called_once_with(fake_file_path, 'r+')

    def test_open_not_existing(self):
        fake_file_path = '/fake_path/file_storage.data'
        fake_dir_name = os.path.dirname(fake_file_path)
        mock_chmod_calls = [
            mock.call(fake_dir_name,
                      stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP),
            mock.call(fake_file_path, stat.S_IRUSR | stat.S_IWUSR)
        ]
        mock_open_calls = [
            mock.call(fake_file_path, 'w'),
            mock.call(fake_file_path, 'r+'),
        ]

        self.mock_object(passwd.os.path, 'isfile', return_value=False)
        self.mock_object(passwd.os.path, 'isdir', return_value=False)
        mock_makedirs = self.mock_object(passwd.os, 'makedirs')
        mock_chmod = self.mock_object(passwd.os, 'chmod')
        mock_open = self.mock_object(
            passwd, 'open', return_value=mock.MagicMock())

        file_storage = passwd.FileStorage(fake_file_path)
        file_storage.open()
        mock_makedirs.assert_called_with(fake_dir_name)
        mock_chmod.assert_has_calls(mock_chmod_calls, any_order=True)
        mock_open.assert_has_calls(mock_open_calls, any_order=True)

    def test_open_not_closed(self):
        fake_file_path = 'file_storage.data'
        fake_file = mock.MagicMock()
        mock_open_calls = [
            mock.call(fake_file_path, 'r+'),
            mock.call(fake_file_path, 'r+'),
        ]
        self.mock_object(passwd.os.path, 'isfile', return_value=True)
        self.mock_object(passwd.os.path, 'isdir', return_value=True)
        mock_open = self.mock_object(passwd, 'open', return_value=fake_file)

        file_storage = passwd.FileStorage(fake_file_path)
        file_storage.open()
        file_storage.open()
        mock_open.assert_has_calls(mock_open_calls)
        fake_file.close.assert_called_once_with()

    def test_load(self):
        passwords = {
            'resource1': {
                'user1': 'resource1-user1',
                'user2': 'resource1-user2',
            },
            'resource2': {
                'user1': 'resource2-user1',
                'user2': 'resource2-user2'
            }
        }
        fake_file_name = 'file_storage.data'
        fake_file_content = json.dumps(passwords)
        fake_file = self._get_fake_file(fake_file_content)
        fake_os_stat = self._get_fake_os_stat(1)

        self._mock_file_open(fake_file, fake_os_stat)

        file_storage = passwd.FileStorage(fake_file_name)
        file_storage.open()
        result = file_storage.load()
        self.assertEqual(passwords, result)

    def test_load_empty_file(self):
        fake_file_name = 'file_storage.data'
        fake_file = self._get_fake_file()
        fake_os_stat = self._get_fake_os_stat(0)

        self._mock_file_open(fake_file, fake_os_stat)

        file_storage = passwd.FileStorage(fake_file_name)
        file_storage.open()
        result = file_storage.load()
        expected = {}
        self.assertEqual(expected, result)

    def test_load_malformed_file(self):
        fake_file_name = 'file_storage.data'
        fake_file = self._get_fake_file('[1, 2, 3]')
        fake_os_stat = self._get_fake_os_stat(1)

        self._mock_file_open(fake_file, fake_os_stat)

        file_storage = passwd.FileStorage(fake_file_name)
        file_storage.open()
        self.assertRaises(ValueError, file_storage.load)

    def test_save(self):
        fake_file_name = 'file_storage.data'
        fake_file = self._get_fake_file('')
        fake_os_stat = self._get_fake_os_stat(0)

        self._mock_file_open(fake_file, fake_os_stat)

        passwords = {
            'resource1': {
                'user1': 'resource1-user1',
                'user2': 'resource1-user2',
            },
            'resource2': {
                'user1': 'resource2-user1',
                'user2': 'resource2-user2'
            }
        }
        fake_file_content = json.dumps(passwords)
        file_storage = passwd.FileStorage(fake_file_name)
        file_storage.open()
        file_storage.save(passwords)
        self.assertEqual(fake_file_content, fake_file.getvalue())

    def test_save_not_dictionary(self):
        fake_file_name = 'file_storage.data'
        fake_file = self._get_fake_file('')
        fake_os_stat = self._get_fake_os_stat(0)

        self._mock_file_open(fake_file, fake_os_stat)

        file_storage = passwd.FileStorage(fake_file_name)
        file_storage.open()
        self.assertRaises(TypeError, file_storage.save, [])

    def test_close(self):
        fake_file_name = 'file_storage.data'
        fake_file = mock.MagicMock()

        self.mock_object(passwd.os.path, 'isfile', return_value=True)
        self.mock_object(passwd.os.path, 'isdir', return_value=True)
        self.mock_object(passwd, 'open', return_value=fake_file)

        file_storage = passwd.FileStorage(fake_file_name)
        file_storage.open()
        file_storage.close()
        fake_file.close.assert_called_once_with()

    def _mock_file_open(self, fake_file, fake_os_stat):
        self.mock_object(passwd.os.path, 'isfile', return_value=True)
        self.mock_object(passwd.os.path, 'isdir', return_value=True)
        self.mock_object(passwd.os, 'stat', return_value=fake_os_stat)
        self.mock_object(passwd, 'open', return_value=fake_file)

    @staticmethod
    def _get_fake_file(content=None):
        return io.StringIO(content)

    @staticmethod
    def _get_fake_os_stat(st_size):
        os_stat = collections.namedtuple('fake_os_stat', ['st_size'])
        os_stat.st_size = st_size
        return os_stat



# Copyright 2011 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""RequestContext: context for requests that persist through all of cinder."""

from __future__ import annotations

import copy
from typing import Any, Optional  # noqa: H301

from keystoneauth1.access import service_catalog as ksa_service_catalog
from keystoneauth1 import plugin
from oslo_config import cfg
from oslo_context import context
from oslo_db.sqlalchemy import enginefacade
from oslo_log import log as logging
from oslo_utils import timeutils

from cinder import exception
from cinder.i18n import _
from cinder.objects import base as objects_base
from cinder import policy

context_opts = [
    cfg.StrOpt('cinder_internal_tenant_project_id',
               help='ID of the project which will be used as the Cinder '
                    'internal tenant.'),
    cfg.StrOpt('cinder_internal_tenant_user_id',
               help='ID of the user to be used in volume operations as the '
                    'Cinder internal tenant.'),
]

CONF = cfg.CONF
CONF.register_opts(context_opts)

LOG = logging.getLogger(__name__)


class _ContextAuthPlugin(plugin.BaseAuthPlugin):
    """A keystoneauth auth plugin that uses the values from the Context.

    Ideally we would use the plugin provided by auth_token middleware however
    this plugin isn't serialized yet so we construct one from the serialized
    auth data.
    """

    def __init__(self, auth_token, sc):
        super(_ContextAuthPlugin, self).__init__()

        self.auth_token = auth_token
        self.service_catalog = ksa_service_catalog.ServiceCatalogV2(sc)

    def get_token(self, *args, **kwargs):
        return self.auth_token

    def get_endpoint(self, session, service_type=None, interface=None,
                     region_name=None, service_name=None, **kwargs):
        return self.service_catalog.url_for(service_type=service_type,
                                            service_name=service_name,
                                            interface=interface,
                                            region_name=region_name)


@enginefacade.transaction_context_provider
class RequestContext(context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """
    def __init__(self,
                 user_id: Optional[str] = None,
                 project_id: Optional[str] = None,
                 is_admin: Optional[bool] = None,
                 read_deleted: Optional[str] = "no",
                 project_name: Optional[str] = None,
                 remote_address: Optional[str] = None,
                 timestamp=None,
                 quota_class=None,
                 service_catalog: Optional[dict] = None,
                 user_auth_plugin=None,
                 message_resource_id = None,
                 message_resource_type = None,
                 message_action = None,
                 **kwargs):
        """Initialize RequestContext.

        :param read_deleted: 'no' indicates deleted records are hidden, 'yes'
            indicates deleted records are visible, 'only' indicates that
            *only* deleted records are visible.

        :param overwrite: Set to False to ensure that the greenthread local
            copy of the index is not overwritten.
        """
        # NOTE(smcginnis): To keep it compatible for code using positional
        # args, explicityly set user_id and project_id in kwargs.
        kwargs.setdefault('user_id', user_id)
        kwargs.setdefault('project_id', project_id)

        super(RequestContext, self).__init__(is_admin=is_admin, **kwargs)

        self.project_name = project_name
        self.read_deleted = read_deleted
        self.remote_address = remote_address
        if not timestamp:
            timestamp = timeutils.utcnow()
        elif isinstance(timestamp, str):
            timestamp = timeutils.parse_isotime(timestamp)
        self.timestamp = timestamp
        self.quota_class = quota_class
        self.message_resource_id = message_resource_id
        self.message_resource_type = message_resource_type
        self.message_action = message_action

        if service_catalog:
            # Only include required parts of service_catalog
            self.service_catalog = [s for s in service_catalog
                                    if s.get('type') in
                                    ('identity', 'compute', 'object-store',
                                     'image', 'key-manager')]
        else:
            # if list is empty or none
            self.service_catalog = []

        # We need to have RequestContext attributes defined
        # when policy.check_is_admin invokes request logging
        # to make it loggable.
        self.is_admin: Optional[bool]
        if self.is_admin is None:
            self.is_admin = policy.check_is_admin(self)
        elif self.is_admin and 'admin' not in self.roles:
            self.roles.append('admin')
        self.user_auth_plugin = user_auth_plugin

    def get_auth_plugin(self):
        if self.user_auth_plugin:
            return self.user_auth_plugin
        else:
            return _ContextAuthPlugin(self.auth_token, self.service_catalog)

    def _get_read_deleted(self) -> str:
        return self._read_deleted

    def _set_read_deleted(self, read_deleted: str) -> None:
        if read_deleted not in ('no', 'yes', 'only'):
            raise ValueError(_("read_deleted can only be one of 'no', "
                               "'yes' or 'only', not %r") % read_deleted)
        self._read_deleted = read_deleted

    def _del_read_deleted(self) -> None:
        del self._read_deleted

    read_deleted = property(_get_read_deleted, _set_read_deleted,
                            _del_read_deleted)

    def to_dict(self) -> dict[str, Any]:
        result = super(RequestContext, self).to_dict()
        result['user_id'] = self.user_id
        result['project_id'] = self.project_id
        result['project_name'] = self.project_name
        result['domain_id'] = self.domain_id
        result['read_deleted'] = self.read_deleted
        result['remote_address'] = self.remote_address
        result['timestamp'] = self.timestamp.isoformat()
        result['quota_class'] = self.quota_class
        result['service_catalog'] = self.service_catalog
        result['request_id'] = self.request_id
        result['message_resource_id'] = self.message_resource_id
        result['message_resource_type'] = self.message_resource_type
        result['message_action'] = self.message_action
        return result

    @classmethod
    def from_dict(cls, values: dict) -> 'RequestContext':
        return cls(user_id=values.get('user_id'),
                   project_id=values.get('project_id'),
                   project_name=values.get('project_name'),
                   domain_id=values.get('domain_id'),
                   read_deleted=values.get('read_deleted'),
                   remote_address=values.get('remote_address'),
                   timestamp=values.get('timestamp'),
                   quota_class=values.get('quota_class'),
                   service_catalog=values.get('service_catalog'),
                   request_id=values.get('request_id'),
                   global_request_id=values.get('global_request_id'),
                   is_admin=values.get('is_admin'),
                   roles=values.get('roles'),
                   auth_token=values.get('auth_token'),
                   user_domain_id=values.get('user_domain_id'),
                   project_domain_id=values.get('project_domain_id'),
                   message_resource_id = values.get('message_resource_id'),
                   message_resource_type = values.get('message_resource_type'),
                   message_action = values.get('message_action')
                   )

    def authorize(self,
                  action: str,
                  target: Optional[dict] = None,
                  target_obj: Optional[dict] = None,
                  fatal: bool = True):
        """Verify that the given action is valid on the target in this context.

        :param action: string representing the action to be checked.
        :param target: dictionary representing the object of the action
            for object creation this should be a dictionary representing the
            location of the object e.g. ``{'project_id': context.project_id}``.
            If None, then this default target will be considered:
            {'project_id': self.project_id, 'user_id': self.user_id}
        :param target_obj: dictionary representing the object which will be
            used to update target.
        :param fatal: if False, will return False when an
            exception.PolicyNotAuthorized occurs.

        :raises cinder.exception.NotAuthorized: if verification fails and fatal
            is True.

        :return: returns a non-False value (not necessarily "True") if
            authorized and False if not authorized and fatal is False.
        """
        if target is None:
            target = {'project_id': self.project_id,
                      'user_id': self.user_id}
        if isinstance(target_obj, objects_base.CinderObject):
            # Turn object into dict so target.update can work
            target.update(
                target_obj.obj_to_primitive()['versioned_object.data'] or {})

            # Ensure 'project_id' and 'user_id' attributes are captured.
            # Some objects (e.g. attachments) have a project_id attribute
            # that isn't present in the dict. The try/except wrappers avoid
            # lazy-load issues when the attribute doesn't exist.
            try:
                target['project_id'] = target_obj.project_id
            except Exception:
                pass
            try:
                target['user_id'] = target_obj.user_id
            except Exception:
                pass
        else:
            target.update(target_obj or {})

        return policy.authorize(self, action, target, do_raise=fatal,
                                exc=exception.PolicyNotAuthorized)

    def to_policy_values(self) -> dict:
        policy = super(RequestContext, self).to_policy_values()

        policy['is_admin'] = self.is_admin

        return policy

    def elevated(self,
                 read_deleted: Optional[str] = None,
                 overwrite: bool = False) -> 'RequestContext':
        """Return a version of this context with admin flag set."""
        context = self.deepcopy()
        context.is_admin = True

        if 'admin' not in context.roles:
            context.roles.append('admin')

        if read_deleted is not None:
            context.read_deleted = read_deleted

        return context

    def deepcopy(self) -> 'RequestContext':
        return copy.deepcopy(self)


def get_admin_context(read_deleted: Optional[str] = "no") -> RequestContext:
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          overwrite=False)


def get_internal_tenant_context() -> Optional[RequestContext]:
    """Build and return the Cinder internal tenant context object

    This request context will only work for internal Cinder operations. It will
    not be able to make requests to remote services. To do so it will need to
    use the keystone client to get an auth_token.
    """
    project_id = CONF.cinder_internal_tenant_project_id
    user_id = CONF.cinder_internal_tenant_user_id

    if project_id and user_id:
        return RequestContext(user_id=user_id,
                              project_id=project_id,
                              is_admin=True,
                              overwrite=False)
    else:
        LOG.warning('Unable to get internal tenant context: Missing '
                    'required config parameters.')
        return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from keystoneauth1 import loading as ks_loading
from keystoneauth1 import service_token
from oslo_config import cfg

from cinder import exception

CONF = cfg.CONF
_SERVICE_AUTH = None

SERVICE_USER_GROUP = 'service_user'

service_user = cfg.OptGroup(
    SERVICE_USER_GROUP,
    title='Service token authentication type options',
    help="""
Configuration options for service to service authentication using a service
token. These options allow to send a service token along with the
user's token when contacting external REST APIs.
"""
)
service_user_opts = [
    cfg.BoolOpt('send_service_user_token',
                default=False,
                help="""
When True, if sending a user token to an REST API, also send a service token.
""")
]

CONF.register_group(service_user)
CONF.register_opts(service_user_opts, group=service_user)

ks_loading.register_session_conf_options(CONF, SERVICE_USER_GROUP)
ks_loading.register_auth_conf_options(CONF, SERVICE_USER_GROUP)


def reset_globals():
    """For async unit test consistency."""
    global _SERVICE_AUTH
    _SERVICE_AUTH = None


def get_service_auth_plugin():
    if CONF.service_user.send_service_user_token:
        global _SERVICE_AUTH
        if not _SERVICE_AUTH:
            _SERVICE_AUTH = ks_loading.load_auth_from_conf_options(
                CONF, group=SERVICE_USER_GROUP)
            if _SERVICE_AUTH is None:
                # This can happen if no auth_type is specified, which probably
                # means there's no auth information in the [service_user] group
                raise exception.ServiceUserTokenNoAuth()
        return _SERVICE_AUTH
    return None


def get_auth_plugin(context, auth=None):
    if auth:
        user_auth = auth
    else:
        user_auth = context.get_auth_plugin()

    service_auth = get_service_auth_plugin()
    if service_auth is not None:
        return service_token.ServiceTokenAuthWrapper(
            user_auth=user_auth, service_auth=service_auth)

    return user_auth



# Stubs for oslo_i18n._translate (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def translate(obj: Any, desired_locale: Optional[Any] = ...): ...



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg
from oslo_log import log as logging

from cinder import exception

CONF = cfg.CONF
CONF.import_group('keystone_authtoken',
                  'keystonemiddleware.auth_token.__init__')

LOG = logging.getLogger(__name__)


def get_volume_type_reservation(ctxt, volume, type_id,
                                reserve_vol_type_only=False,
                                negative=False):
    from cinder import quota
    QUOTAS = quota.QUOTAS
    # Reserve quotas for the given volume type
    try:
        reserve_opts = {'volumes': 1, 'gigabytes': volume['size']}
        # When retyping a volume it may contain snapshots (if we are not
        # migrating it) and we need to account for its snapshots' size
        if volume.snapshots:
            reserve_opts['snapshots'] = len(volume.snapshots)
            if not CONF.no_snapshot_gb_quota:
                reserve_opts['gigabytes'] += sum(snap.volume_size
                                                 for snap in volume.snapshots)
        QUOTAS.add_volume_type_opts(ctxt,
                                    reserve_opts,
                                    type_id)
        # If reserve_vol_type_only is True, just reserve volume_type quota,
        # not volume quota.
        if reserve_vol_type_only:
            reserve_opts.pop('volumes')
            reserve_opts.pop('gigabytes')
            reserve_opts.pop('snapshots', None)

        if negative:
            for key, value in reserve_opts.items():
                reserve_opts[key] = -value

        # Note that usually the project_id on the volume will be the same as
        # the project_id in the context. But, if they are different then the
        # reservations must be recorded against the project_id that owns the
        # volume.
        project_id = volume['project_id']
        reservations = QUOTAS.reserve(ctxt,
                                      project_id=project_id,
                                      **reserve_opts)
    except exception.OverQuota as e:
        process_reserve_over_quota(ctxt, e,
                                   resource='volumes',
                                   size=volume.size)
    return reservations


def _filter_domain_id_from_parents(domain_id, tree):
    """Removes the domain_id from the tree if present"""
    new_tree = None
    if tree:
        parent, children = next(iter(tree.items()))
        # Don't add the domain id to the parents hierarchy
        if parent != domain_id:
            new_tree = {parent: _filter_domain_id_from_parents(domain_id,
                                                               children)}

    return new_tree


OVER_QUOTA_RESOURCE_EXCEPTIONS = {'snapshots': exception.SnapshotLimitExceeded,
                                  'backups': exception.BackupLimitExceeded,
                                  'volumes': exception.VolumeLimitExceeded,
                                  'groups': exception.GroupLimitExceeded}


def process_reserve_over_quota(context, over_quota_exception,
                               resource, size=None):
    """Handle OverQuota exception.

    Analyze OverQuota exception, and raise new exception related to
    resource type. If there are unexpected items in overs,
    UnexpectedOverQuota is raised.

    :param context: security context
    :param over_quota_exception: OverQuota exception
    :param resource: can be backups, snapshots, and volumes
    :param size: requested size in reservation
    """
    def _consumed(name):
        return (usages[name]['reserved'] + usages[name]['in_use'])

    overs = over_quota_exception.kwargs['overs']
    usages = over_quota_exception.kwargs['usages']
    quotas = over_quota_exception.kwargs['quotas']
    invalid_overs = []

    for over in overs:
        if 'gigabytes' in over:
            msg = ("Quota exceeded for %(s_pid)s, tried to create "
                   "%(s_size)dG %(s_resource)s (%(d_consumed)dG of "
                   "%(d_quota)dG already consumed).")
            LOG.warning(msg, {'s_pid': context.project_id,
                              's_size': size,
                              's_resource': resource[:-1],
                              'd_consumed': _consumed(over),
                              'd_quota': quotas[over]})
            if resource == 'backups':
                exc = exception.VolumeBackupSizeExceedsAvailableQuota
            else:
                exc = exception.VolumeSizeExceedsAvailableQuota
            raise exc(
                name=over,
                requested=size,
                consumed=_consumed(over),
                quota=quotas[over])
        if (resource in OVER_QUOTA_RESOURCE_EXCEPTIONS.keys() and
                resource in over):
            msg = ("Quota exceeded for %(s_pid)s, tried to create "
                   "%(s_resource)s (%(d_consumed)d %(s_resource)ss "
                   "already consumed).")
            LOG.warning(msg, {'s_pid': context.project_id,
                              'd_consumed': _consumed(over),
                              's_resource': resource[:-1]})
            raise OVER_QUOTA_RESOURCE_EXCEPTIONS[resource](
                allowed=quotas[over],
                name=over)
        invalid_overs.append(over)

    if invalid_overs:
        raise exception.UnexpectedOverQuota(name=', '.join(invalid_overs))



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Generic Node base class for all workers that run on hosts."""


import inspect
import os
import random
import subprocess
import sys
import time
from typing import Optional

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import service
from oslo_service import wsgi
from oslo_utils import importutils
osprofiler_initializer = importutils.try_import('osprofiler.initializer')
profiler = importutils.try_import('osprofiler.profiler')
profiler_opts = importutils.try_import('osprofiler.opts')


from cinder.common import constants
from cinder import context
from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base as objects_base
from cinder.objects import fields
from cinder import rpc
from cinder import version
from cinder.volume import volume_utils

if os.name == 'nt':
    from os_win import utilsfactory as os_win_utilsfactory
else:
    os_win_utilsfactory = None


LOG = logging.getLogger(__name__)

service_opts = [
    cfg.IntOpt('report_interval',
               default=10,
               help='Interval, in seconds, between nodes reporting state '
                    'to datastore'),
    cfg.IntOpt('periodic_interval',
               default=60,
               help='Interval, in seconds, between running periodic tasks'),
    cfg.IntOpt('periodic_fuzzy_delay',
               default=60,
               help='Range, in seconds, to randomly delay when starting the'
                    ' periodic task scheduler to reduce stampeding.'
                    ' (Disable by setting to 0)'),
    cfg.StrOpt('osapi_volume_listen',
               default="0.0.0.0",
               help='IP address on which OpenStack Volume API listens'),
    cfg.PortOpt('osapi_volume_listen_port',
                default=8776,
                help='Port on which OpenStack Volume API listens'),
    cfg.IntOpt('osapi_volume_workers',
               help='Number of workers for OpenStack Volume API service. '
                    'The default is equal to the number of CPUs available.'),
    cfg.BoolOpt('osapi_volume_use_ssl',
                default=False,
                help='Wraps the socket in a SSL context if True is set. '
                     'A certificate file and key file must be specified.'), ]


CONF = cfg.CONF
CONF.register_opts(service_opts)
if profiler_opts:
    profiler_opts.set_defaults(CONF)


def setup_profiler(binary: str, host: str) -> None:
    if (osprofiler_initializer is None or
            profiler is None or
            profiler_opts is None):
        LOG.debug('osprofiler is not present')
        return

    if CONF.profiler.enabled:
        osprofiler_initializer.init_from_conf(
            conf=CONF,
            context=context.get_admin_context().to_dict(),
            project="cinder",
            service=binary,
            host=host
        )
        LOG.warning(
            "OSProfiler is enabled.\nIt means that person who knows "
            "any of hmac_keys that are specified in "
            "/etc/cinder/cinder.conf can trace his requests. \n"
            "In real life only operator can read this file so there "
            "is no security issue. Note that even if person can "
            "trigger profiler, only admin user can retrieve trace "
            "information.\n"
            "To disable OSProfiler set in cinder.conf:\n"
            "[profiler]\nenabled=false")


class Service(service.Service):
    """Service object for binaries running on hosts.

    A service takes a manager and enables rpc by listening to queues based
    on topic. It also periodically runs tasks on the manager and reports
    it state to the database services table.
    """
    # Make service_id a class attribute so it can be used for clean up
    service_id = None

    def __init__(self, host: str, binary: str, topic: str,
                 manager: str,
                 report_interval: Optional[int] = None,
                 periodic_interval: Optional[int] = None,
                 periodic_fuzzy_delay: Optional[int] = None,
                 service_name: Optional[str] = None,
                 coordination: bool = False,
                 cluster: Optional[str] = None,
                 *args, **kwargs):
        super(Service, self).__init__()

        if not rpc.initialized():
            rpc.init(CONF)

        self.cluster = cluster
        self.host = host
        self.binary = binary
        self.topic = topic
        self.manager_class_name = manager
        self.coordination = coordination
        manager_class = importutils.import_class(self.manager_class_name)
        if CONF.profiler.enabled:
            manager_class = profiler.trace_cls("rpc")(manager_class)

        self.service = None
        self.manager = manager_class(host=self.host,
                                     cluster=self.cluster,
                                     service_name=service_name,
                                     *args, **kwargs)
        self.availability_zone: str = self.manager.availability_zone
        self.model_disconnected: bool

        # NOTE(geguileo): We need to create the Service DB entry before we
        # create the manager, otherwise capped versions for serializer and rpc
        # client would use existing DB entries not including us, which could
        # result in us using None (if it's the first time the service is run)
        # or an old version (if this is a normal upgrade of a single service).
        ctxt = context.get_admin_context()
        try:
            service_ref = objects.Service.get_by_args(ctxt, host, binary)
            service_ref.rpc_current_version = manager_class.RPC_API_VERSION
            obj_version = objects_base.OBJ_VERSIONS.get_current()
            service_ref.object_current_version = obj_version

            # added_to_cluster attribute marks when we consider that we have
            # just added a host to a cluster so we can include resources into
            # that cluster.  We consider that we have added the host when we
            # didn't have data in the cluster DB field and our current
            # configuration has a cluster value.  We don't want to do anything
            # automatic if the cluster is changed, in those cases we'll want
            # to use cinder manage command and to it manually.
            self.added_to_cluster = (not service_ref.cluster_name and cluster)

            if service_ref.cluster_name != cluster:
                LOG.info('This service has been moved from cluster '
                         '%(cluster_svc)s to %(cluster_cfg)s. Resources '
                         'will %(opt_no)sbe moved to the new cluster',
                         {'cluster_svc': service_ref.cluster_name,
                          'cluster_cfg': cluster,
                          'opt_no': '' if self.added_to_cluster else 'NOT '})

            if self.added_to_cluster:
                # We pass copy service's disable status in the cluster if we
                # have to create it.
                self._ensure_cluster_exists(ctxt, service_ref)
                service_ref.cluster_name = cluster
            service_ref.save()
            Service.service_id = service_ref.id
            self.origin_service_id = service_ref.id
        except exception.NotFound:
            self._create_service_ref(ctxt, manager_class.RPC_API_VERSION)
            # Service entry Entry didn't exist because it was manually removed
            # or it's the first time running, to be on the safe side we say we
            # were added if we are clustered.
            self.added_to_cluster = bool(cluster)  # type: ignore

        self.report_interval = report_interval
        self.periodic_interval = periodic_interval
        self.periodic_fuzzy_delay = periodic_fuzzy_delay
        self.basic_config_check()
        self.saved_args, self.saved_kwargs = args, kwargs

        setup_profiler(binary, host)
        self.rpcserver: Optional['messaging.rpc.RPCServer'] = None
        self.backend_rpcserver: Optional['messaging.rpc.RPCServer'] = None
        self.cluster_rpcserver: Optional['messaging.rpc.RPCServer'] = None

    def start(self) -> None:
        version_string = version.version_string()
        LOG.info('Starting %(topic)s node (version %(version_string)s)',
                 {'topic': self.topic, 'version_string': version_string})
        self.model_disconnected = False

        if self.coordination:
            coordination.COORDINATOR.start()

        # NOTE(yikun): When re-spawning child process, we should set the class
        # attribute back using the origin service_id, otherwise,
        # the Service.service_id will be inherited from the parent process,
        # and will be recorded as the last started service id by mistaken.
        Service.service_id = self.origin_service_id

        self.manager.init_host(added_to_cluster=self.added_to_cluster,
                               service_id=Service.service_id)

        LOG.debug("Creating RPC server for service %s", self.topic)

        ctxt = context.get_admin_context()
        endpoints = [self.manager]
        endpoints.extend(self.manager.additional_endpoints)
        obj_version_cap = objects.Service.get_minimum_obj_version(ctxt)
        LOG.debug("Pinning object versions for RPC server serializer to %s",
                  obj_version_cap)
        serializer = objects_base.CinderObjectSerializer(obj_version_cap)

        target = messaging.Target(topic=self.topic, server=self.host)
        self.rpcserver = rpc.get_server(target, endpoints, serializer)
        self.rpcserver.start()

        # NOTE(dulek): Kids, don't do that at home. We're relying here on
        # oslo.messaging implementation details to keep backward compatibility
        # with pre-Ocata services. This will not matter once we drop
        # compatibility with them.
        if self.topic == constants.VOLUME_TOPIC:
            target = messaging.Target(
                topic='%(topic)s.%(host)s' % {'topic': self.topic,
                                              'host': self.host},
                server=volume_utils.extract_host(self.host, 'host'))
            self.backend_rpcserver = rpc.get_server(target, endpoints,
                                                    serializer)
            self.backend_rpcserver.start()

        if self.cluster:
            LOG.info('Starting %(topic)s cluster %(cluster)s (version '
                     '%(version)s)',
                     {'topic': self.topic, 'version': version_string,
                      'cluster': self.cluster})
            target = messaging.Target(
                topic='%s.%s' % (self.topic, self.cluster),
                server=volume_utils.extract_host(self.cluster, 'host'))
            serializer = objects_base.CinderObjectSerializer(obj_version_cap)
            self.cluster_rpcserver = rpc.get_server(target, endpoints,
                                                    serializer)
            self.cluster_rpcserver.start()

        self.manager.init_host_with_rpc()

        if self.report_interval:
            self.tg.add_timer(self.report_interval, self.report_state,
                              initial_delay=self.report_interval)

        if self.periodic_interval:
            initial_delay: Optional[int]
            if self.periodic_fuzzy_delay:
                initial_delay = random.randint(0, self.periodic_fuzzy_delay)
            else:
                initial_delay = None
            self.tg.add_timer(self.periodic_interval, self.periodic_tasks,
                              initial_delay=initial_delay)

    def basic_config_check(self) -> None:
        """Perform basic config checks before starting service."""
        # Make sure report interval is less than service down time
        if self.report_interval:
            if CONF.service_down_time <= self.report_interval:
                new_down_time = int(self.report_interval * 2.5)
                LOG.warning(
                    "Report interval must be less than service down "
                    "time. Current config service_down_time: "
                    "%(service_down_time)s, report_interval for this: "
                    "service is: %(report_interval)s. Setting global "
                    "service_down_time to: %(new_down_time)s",
                    {'service_down_time': CONF.service_down_time,
                     'report_interval': self.report_interval,
                     'new_down_time': new_down_time})
                CONF.set_override('service_down_time', new_down_time)

    def _ensure_cluster_exists(self,
                               context: context.RequestContext,
                               service: 'Service') -> None:
        if self.cluster:
            try:
                cluster = objects.Cluster.get_by_id(context, None,
                                                    name=self.cluster,
                                                    binary=self.binary)
                # If the cluster already exists, then the service replication
                # fields must match those of the cluster unless the service
                # is in error status.
                error_states = (fields.ReplicationStatus.ERROR,
                                fields.ReplicationStatus.FAILOVER_ERROR)
                if service.replication_status not in error_states:
                    for attr in ('replication_status', 'active_backend_id',
                                 'frozen'):
                        if getattr(service, attr) != getattr(cluster, attr):
                            setattr(service, attr, getattr(cluster, attr))

            except exception.ClusterNotFound:
                # Since the cluster didn't exist, we copy replication fields
                # from the service.
                cluster = objects.Cluster(
                    context=context,
                    name=self.cluster,
                    binary=self.binary,
                    disabled=service.disabled,
                    replication_status=service.replication_status,
                    active_backend_id=service.active_backend_id,
                    frozen=service.frozen)
                try:
                    cluster.create()

                # Race condition occurred and another service created the
                # cluster, so we can continue as it already exists.
                except exception.ClusterExists:
                    pass

    def _create_service_ref(self,
                            context: context.RequestContext,
                            rpc_version: Optional[str] = None) -> None:
        kwargs = {
            'host': self.host,
            'binary': self.binary,
            'topic': self.topic,
            'report_count': 0,
            'availability_zone': self.availability_zone,
            'rpc_current_version': rpc_version or self.manager.RPC_API_VERSION,
            'object_current_version': objects_base.OBJ_VERSIONS.get_current(),
        }
        kwargs['cluster_name'] = self.cluster
        service_ref = objects.Service(context=context, **kwargs)
        service_ref.create()
        Service.service_id = service_ref.id
        self.origin_service_id = service_ref.id
        self._ensure_cluster_exists(context, service_ref)
        # If we have updated the service_ref with replication data from
        # the cluster it will be saved.
        service_ref.save()

    def __getattr__(self, key: str):
        manager = self.__dict__.get('manager', None)
        return getattr(manager, key)

    @classmethod
    def create(cls,
               host: Optional[str] = None,
               binary: Optional[str] = None,
               topic: Optional[str] = None,
               manager: Optional[str] = None,
               report_interval: Optional[int] = None,
               periodic_interval: Optional[int] = None,
               periodic_fuzzy_delay: Optional[int] = None,
               service_name: Optional[str] = None,
               coordination: bool = False,
               cluster: Optional[str] = None,
               **kwargs) -> 'Service':
        """Instantiates class and passes back application object.

        :param host: defaults to CONF.host
        :param binary: defaults to basename of executable
        :param topic: defaults to bin_name - 'cinder-' part
        :param manager: defaults to CONF.<topic>_manager
        :param report_interval: defaults to CONF.report_interval
        :param periodic_interval: defaults to CONF.periodic_interval
        :param periodic_fuzzy_delay: defaults to CONF.periodic_fuzzy_delay
        :param cluster: Defaults to None, as only some services will have it

        """
        if not host:
            host = CONF.host
        if not binary:
            binary = os.path.basename(inspect.stack()[-1][1])
        if not topic:
            topic = binary
        if not manager:
            subtopic = topic.rpartition('cinder-')[2]
            manager = CONF.get('%s_manager' % subtopic, None)
        if report_interval is None:
            report_interval = CONF.report_interval
        if periodic_interval is None:
            periodic_interval = CONF.periodic_interval
        if periodic_fuzzy_delay is None:
            periodic_fuzzy_delay = CONF.periodic_fuzzy_delay

        assert host is not None
        assert manager is not None
        service_obj = cls(host, binary, topic, manager,
                          report_interval=report_interval,
                          periodic_interval=periodic_interval,
                          periodic_fuzzy_delay=periodic_fuzzy_delay,
                          service_name=service_name,
                          coordination=coordination,
                          cluster=cluster, **kwargs)

        return service_obj

    def stop(self) -> None:
        # Try to shut the connection down, but if we get any sort of
        # errors, go ahead and ignore them.. as we're shutting down anyway
        try:
            if self.rpcserver is not None:
                self.rpcserver.stop()
            if self.backend_rpcserver:
                self.backend_rpcserver.stop()
            if self.cluster_rpcserver:
                self.cluster_rpcserver.stop()
        except Exception:
            pass

        if self.coordination:
            try:
                coordination.COORDINATOR.stop()
            except Exception:
                pass
        super(Service, self).stop(graceful=True)

    def wait(self) -> None:
        if self.rpcserver:
            self.rpcserver.wait()
        if self.backend_rpcserver:
            self.backend_rpcserver.wait()
        if self.cluster_rpcserver:
            self.cluster_rpcserver.wait()
        super(Service, self).wait()

    def periodic_tasks(self, raise_on_error: bool = False) -> None:
        """Tasks to be run at a periodic interval."""
        ctxt = context.get_admin_context()
        self.manager.run_periodic_tasks(ctxt, raise_on_error=raise_on_error)

    def report_state(self) -> None:
        """Update the state of this service in the datastore."""
        if not self.manager.is_working():
            # NOTE(dulek): If manager reports a problem we're not sending
            # heartbeats - to indicate that service is actually down.
            LOG.error('Manager for service %(binary)s %(host)s is '
                      'reporting problems, not sending heartbeat. '
                      'Service will appear "down".',
                      {'binary': self.binary,
                       'host': self.host})
            return

        ctxt = context.get_admin_context()
        try:
            try:
                service_ref = objects.Service.get_by_id(ctxt,
                                                        Service.service_id)
            except exception.NotFound:
                LOG.debug('The service database object disappeared, '
                          'recreating it.')
                self._create_service_ref(ctxt)
                service_ref = objects.Service.get_by_id(ctxt,
                                                        Service.service_id)

            service_ref.report_count += 1
            if self.availability_zone != service_ref.availability_zone:
                service_ref.availability_zone = self.availability_zone

            service_ref.save()

            # TODO(termie): make this pattern be more elegant.
            if getattr(self, 'model_disconnected', False):
                self.model_disconnected = False
                LOG.error('Recovered model server connection!')

        except db_exc.DBConnectionError:
            if not getattr(self, 'model_disconnected', False):
                self.model_disconnected = True
                LOG.exception('model server went away')

        # NOTE(jsbryant) Other DB errors can happen in HA configurations.
        # such errors shouldn't kill this thread, so we handle them here.
        except db_exc.DBError:
            if not getattr(self, 'model_disconnected', False):
                self.model_disconnected = True
                LOG.exception('DBError encountered: ')

        except Exception:
            if not getattr(self, 'model_disconnected', False):
                self.model_disconnected = True
                LOG.exception('Exception encountered: ')

    def reset(self) -> None:
        self.manager.reset()
        super(Service, self).reset()


class WSGIService(service.ServiceBase):
    """Provides ability to launch API from a 'paste' configuration."""

    def __init__(self, name, loader=None):
        """Initialize, but do not start the WSGI server.

        :param name: The name of the WSGI server given to the loader.
        :param loader: Loads the WSGI application using the given name.
        :returns: None

        """
        self.name = name
        self.manager = self._get_manager()
        self.loader = loader or wsgi.Loader(CONF)
        self.app = self.loader.load_app(name)
        self.host = getattr(CONF, '%s_listen' % name, "0.0.0.0")
        self.port = getattr(CONF, '%s_listen_port' % name, 0)
        self.use_ssl = getattr(CONF, '%s_use_ssl' % name, False)
        self.workers = (getattr(CONF, '%s_workers' % name, None) or
                        processutils.get_worker_count())
        if self.workers and self.workers < 1:
            worker_name = '%s_workers' % name
            msg = (_("%(worker_name)s value of %(workers)d is invalid, "
                     "must be greater than 0.") %
                   {'worker_name': worker_name,
                    'workers': self.workers})
            raise exception.InvalidConfigurationValue(msg)
        setup_profiler(name, self.host)

        self.server = wsgi.Server(CONF,
                                  name,
                                  self.app,
                                  host=self.host,
                                  port=self.port,
                                  use_ssl=self.use_ssl)

    def _get_manager(self):
        """Initialize a Manager object appropriate for this service.

        Use the service name to look up a Manager subclass from the
        configuration and initialize an instance. If no class name
        is configured, just return None.

        :returns: a Manager instance, or None.

        """
        fl = '%s_manager' % self.name
        if fl not in CONF:
            return None

        manager_class_name = CONF.get(fl, None)
        if not manager_class_name:
            return None

        manager_class = importutils.import_class(manager_class_name)
        return manager_class()

    def start(self) -> None:
        """Start serving this service using loaded configuration.

        Also, retrieve updated port number in case '0' was passed in, which
        indicates a random port should be used.

        :returns: None

        """
        if self.manager:
            self.manager.init_host()
        self.server.start()
        self.port = self.server.port

    def stop(self) -> None:
        """Stop serving this API.

        :returns: None

        """
        self.server.stop()

    def wait(self) -> None:
        """Wait for the service to stop serving this API.

        :returns: None

        """
        self.server.wait()

    def reset(self) -> None:
        """Reset server greenpool size to default.

        :returns: None

        """
        self.server.reset()


def process_launcher() -> service.ProcessLauncher:
    return service.ProcessLauncher(CONF, restart_method='mutate')


# NOTE(vish): the global launcher is to maintain the existing
#             functionality of calling service.serve +
#             service.wait
_launcher = None


def serve(server, workers=None):
    global _launcher
    if _launcher:
        raise RuntimeError(_('serve() can only be called once'))

    _launcher = service.launch(CONF, server, workers=workers,
                               restart_method='mutate')


def wait() -> None:
    CONF.log_opt_values(LOG, logging.DEBUG)

    try:
        _launcher.wait()  # type: ignore
    except KeyboardInterrupt:
        _launcher.stop()  # type: ignore
    rpc.cleanup()


class Launcher(object):
    def __init__(self):
        self.launch_service = serve
        self.wait = wait


def get_launcher() -> service.ProcessLauncher:
    # Note(lpetrut): ProcessLauncher uses green pipes which fail on Windows
    # due to missing support of non-blocking I/O pipes. For this reason, the
    # service must be spawned differently on Windows, using the ServiceLauncher
    # class instead.
    if os.name == 'nt':
        return Launcher()
    else:
        return process_launcher()


class WindowsProcessLauncher(object):
    def __init__(self):
        self._processutils = os_win_utilsfactory.get_processutils()

        self._workers = []
        self._worker_job_handles = []
        self._signal_handler = service.SignalHandler()
        self._add_signal_handlers()

    def add_process(self, cmd):
        LOG.info("Starting subprocess: %s", cmd)

        worker = subprocess.Popen(cmd)
        try:
            job_handle = self._processutils.kill_process_on_job_close(
                worker.pid)
        except Exception:
            LOG.exception("Could not associate child process "
                          "with a job, killing it.")
            worker.kill()
            raise

        self._worker_job_handles.append(job_handle)
        self._workers.append(worker)

    def _add_signal_handlers(self):
        self._signal_handler.add_handler('SIGINT', self._terminate)
        self._signal_handler.add_handler('SIGTERM', self._terminate)

    def _terminate(self, *args):
        # We've already assigned win32 job objects to child processes,
        # requesting them to stop once all the job handles are closed.
        # When this process dies, so will the child processes.
        LOG.info("Received request to terminate.")
        sys.exit(1)

    def wait(self):
        pids = [worker.pid for worker in self._workers]
        if pids:
            self._processutils.wait_for_multiple_processes(pids,
                                                           wait_all=True)
        # By sleeping here, we allow signal handlers to be executed.
        time.sleep(0)



# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from cinder.tests.functional.api import client
from cinder.tests.functional import functional_helpers
from cinder.volume import configuration


class VolumesTest(functional_helpers._FunctionalTestBase):
    _vol_type_name = 'functional_test_type'

    def setUp(self):
        super(VolumesTest, self).setUp()
        self.api.create_type(self._vol_type_name)

    def _get_flags(self):
        f = super(VolumesTest, self)._get_flags()
        f['volume_driver'] = (
            {'v': 'cinder.tests.fake_driver.FakeLoggingVolumeDriver',
             'g': configuration.SHARED_CONF_GROUP})
        f['default_volume_type'] = {'v': self._vol_type_name}
        return f

    def test_get_volumes_summary(self):
        """Simple check that listing volumes works."""
        volumes = self.api.get_volumes(False)
        self.assertIsNotNone(volumes)

    def test_get_volumes(self):
        """Simple check that listing volumes works."""
        volumes = self.api.get_volumes()
        self.assertIsNotNone(volumes)

    def test_create_and_delete_volume(self):
        """Creates and deletes a volume."""

        # Create volume
        created_volume = self.api.post_volume({'volume': {'size': 1}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(self._vol_type_name, found_volume['volume_type'])

        # It should also be in the all-volume list
        volumes = self.api.get_volumes()
        volume_names = [volume['id'] for volume in volumes]
        self.assertIn(created_volume_id, volume_names)

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])

        # It should be available...
        self.assertEqual('available', found_volume['status'])

        # Delete the volume
        self.api.delete_volume(created_volume_id)

        # Wait (briefly) for deletion. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])

        # Should be gone
        self.assertIsNone(found_volume)

    def test_create_no_volume_type(self):
        """Verify volume_type is not None"""

        # Create volume
        created_volume = self.api.post_volume({'volume': {'size': 1}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])
        self.assertEqual('available', found_volume['status'])

        # It should have a volume_type
        self.assertIsNotNone(found_volume['volume_type'])

        # Delete the volume
        self.api.delete_volume(created_volume_id)
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        self.assertIsNone(found_volume)

    def test_create_volume_default_type(self):
        """Verify that the configured default_volume_type is used"""

        my_vol_type_name = 'default_type'
        self.api.create_type(my_vol_type_name)
        self.flags(default_volume_type=my_vol_type_name)

        # Create volume
        created_volume = self.api.post_volume({'volume': {'size': 1}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])
        self.assertEqual('available', found_volume['status'])

        # It should have the default volume_type
        self.assertEqual(my_vol_type_name, found_volume['volume_type'])

        # Delete the volume
        self.api.delete_volume(created_volume_id)
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        self.assertIsNone(found_volume)

    def test_create_volume_bad_default_type(self):
        """Verify non-existent default volume type errors out."""

        # configure a non-existent default type
        self.flags(default_volume_type='non-existent-type')

        # Create volume and verify it errors out with 500 status
        self.assertRaises(client.OpenStackApiException500,
                          self.api.post_volume, {'volume': {'size': 1}})

    def test_create_volume_default_type_set_none(self):
        """Verify None default volume type errors out."""

        # configure None default type
        self.flags(default_volume_type=None)

        # Create volume and verify it errors out with 500 status
        self.assertRaises(client.OpenStackApiException500,
                          self.api.post_volume, {'volume': {'size': 1}})

    @mock.patch('cinder.api.common.get_cluster_host',
                return_value=(None, None))
    def test_manage_volume_default_type_set_none(self, fake_get_host):
        """Verify missing default volume type errors out when managing."""

        # configure None default type
        self.flags(default_volume_type=None)

        # manage something in the backend and verify you get a 500
        self.assertRaises(client.OpenStackApiException500,
                          self.api.post_manage_volume)

        # make sure that we actually made it into the method we
        # want to test and the 500 wasn't from something else
        fake_get_host.assert_called_once()

    def test_create_volume_specified_type(self):
        """Verify volume_type is not default."""

        my_vol_type_name = 'my_specified_type'
        my_vol_type_id = self.api.create_type(my_vol_type_name)['id']

        # Create volume
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'volume_type': my_vol_type_id}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        found_volume = self._poll_volume_while(created_volume_id, ['creating'])
        self.assertEqual('available', found_volume['status'])

        # It should have the specified volume_type
        self.assertEqual(my_vol_type_name, found_volume['volume_type'])

        # Delete the volume and test type
        self.api.delete_volume(created_volume_id)
        found_volume = self._poll_volume_while(created_volume_id, ['deleting'])
        self.assertIsNone(found_volume)
        self.api.delete_type(my_vol_type_id)

    def test_create_volume_from_source_vol_inherits_voltype(self):
        src_vol_type_name = 'source_vol_type'
        src_vol_type_id = self.api.create_type(src_vol_type_name)['id']

        # Create source volume
        src_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'volume_type': src_vol_type_id}})
        self.assertTrue(uuidutils.is_uuid_like(src_volume['id']))
        src_volume_id = src_volume['id']

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        src_volume = self._poll_volume_while(src_volume_id, ['creating'])
        self.assertEqual('available', src_volume['status'])

        # Create a new volume using src_volume, do not specify a volume_type
        new_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'source_volid': src_volume_id}})
        new_volume_id = new_volume['id']

        # Wait for creation ...
        new_volume = self._poll_volume_while(new_volume_id, ['creating'])
        self.assertEqual('available', new_volume['status'])

        # It should have the same type as the source volume
        self.assertEqual(src_vol_type_name, new_volume['volume_type'])

        # Delete the volumes and test type
        self.api.delete_volume(src_volume_id)
        found_volume = self._poll_volume_while(src_volume_id, ['deleting'])
        self.assertIsNone(found_volume)
        self.api.delete_volume(new_volume_id)
        found_volume = self._poll_volume_while(new_volume_id, ['deleting'])
        self.assertIsNone(found_volume)
        self.api.delete_type(src_vol_type_id)

    def test_create_volume_from_snapshot_inherits_voltype(self):
        src_vol_type_name = 'a_very_new_vol_type'
        src_vol_type_id = self.api.create_type(src_vol_type_name)['id']

        # Create source volume
        src_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'volume_type': src_vol_type_id}})
        src_volume_id = src_volume['id']

        # Wait (briefly) for creation. Delay is due to the 'message queue'
        src_volume = self._poll_volume_while(src_volume_id, ['creating'])
        self.assertEqual('available', src_volume['status'])

        # Create a snapshot of src_volume
        snapshot = self.api.post_snapshot(
            {'snapshot': {'volume_id': src_volume_id,
                          'name': 'test_snapshot'}})
        self.assertEqual(src_volume_id, snapshot['volume_id'])
        snapshot_id = snapshot['id']

        # make sure the snapshot is ready
        snapshot = self._poll_snapshot_while(snapshot_id, ['creating'])
        self.assertEqual('available', snapshot['status'])

        # create a new volume from the snapshot, do not specify a volume_type
        new_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'snapshot_id': snapshot_id}})
        new_volume_id = new_volume['id']

        # Wait for creation ...
        new_volume = self._poll_volume_while(new_volume_id, ['creating'])
        self.assertEqual('available', new_volume['status'])

        # Finally, here's the whole point of this test:
        self.assertEqual(src_vol_type_name, new_volume['volume_type'])

        # Delete the snapshot, volumes, and test type
        self.api.delete_snapshot(snapshot_id)
        snapshot = self._poll_snapshot_while(snapshot_id, ['deleting'])
        self.assertIsNone(snapshot)

        self.api.delete_volume(src_volume_id)
        src_volume = self._poll_volume_while(src_volume_id, ['deleting'])
        self.assertIsNone(src_volume)

        self.api.delete_volume(new_volume_id)
        new_volume = self._poll_volume_while(new_volume_id, ['deleting'])
        self.assertIsNone(new_volume)

        self.api.delete_type(src_vol_type_id)

    def test_create_volume_with_metadata(self):
        """Creates a volume with metadata."""

        # Create volume
        metadata = {'key1': 'value1',
                    'key2': 'value2',
                    'volume/created/by': 'cinder'}
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'metadata': metadata}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there and metadata present
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(metadata, found_volume['metadata'])

    def test_create_volume_in_availability_zone(self):
        """Creates a volume in availability_zone."""

        # Create volume
        availability_zone = 'nova'
        created_volume = self.api.post_volume(
            {'volume': {'size': 1,
                        'availability_zone': availability_zone}})
        self.assertTrue(uuidutils.is_uuid_like(created_volume['id']))
        created_volume_id = created_volume['id']

        # Check it's there and availability zone present
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual(availability_zone, found_volume['availability_zone'])

    def test_create_and_update_volume(self):
        # Create vol1
        created_volume = self.api.post_volume({'volume': {
            'size': 1, 'name': 'vol1'}})
        self.assertEqual('vol1', created_volume['name'])
        created_volume_id = created_volume['id']

        # update volume
        body = {'volume': {'name': 'vol-one'}}
        updated_volume = self.api.put_volume(created_volume_id, body)
        self.assertEqual('vol-one', updated_volume['name'])

        # check for update
        found_volume = self.api.get_volume(created_volume_id)
        self.assertEqual(created_volume_id, found_volume['id'])
        self.assertEqual('vol-one', found_volume['name'])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock

import ddt

from cinder.api.contrib import volume_image_metadata as image_metadata
from cinder.api import microversions as mv
from cinder.api.v3 import volume_metadata
from cinder import db
from cinder import exception
from cinder.policies import volume_metadata as policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base
from cinder.tests.unit.policies import test_base
from cinder.tests.unit import utils as test_utils
from cinder.volume import api as volume_api


@ddt.ddt
class VolumeMetadataPolicyTest(base.BasePolicyTest):
    authorized_readers = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_readers = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_members = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # DB validations will throw VolumeNotFound for some contexts
    unauthorized_exceptions = [
        exception.VolumeNotFound,
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = volume_metadata.Controller()
        self.image_controller = image_metadata.VolumeImageMetadataController()
        self.api_path = '/v3/%s/volumes' % (self.project_id)
        self.api_version = mv.BASE_VERSION

    def _create_volume(self, image_metadata=None, **kwargs):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 name='fake_vol_type',
                                                 testcase_instance=self)

        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=vol_type.id,
                                          testcase_instance=self, **kwargs)

        for (k, v) in (image_metadata.items() if image_metadata else []):
            db.volume_glance_metadata_create(self.project_admin_context,
                                             volume.id, k, v)
        return volume

    @ddt.data(*base.all_users)
    def test_get_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.GET_POLICY
        url = '%s/%s/metadata' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        self.common_policy_check(user_id, self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.index, req,
                                 volume_id=volume.id)

    @ddt.data(*base.all_users)
    def test_create_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.CREATE_POLICY
        url = '%s/%s/metadata' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "metadata": {
                "name": "metadata0"
            }
        }

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 volume_id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_update_policy(self, user_id):
        volume = self._create_volume(metadata={"foo": "bar"})
        rule_name = policy.UPDATE_POLICY
        url = '%s/%s/metadata' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'PUT'
        body = {
            # Not sure why, but the API code expects the body to contain
            # a "meta" (not "metadata") dict.
            "meta": {
                "foo": "zap"
            }
        }

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.update, req,
                                 volume_id=volume.id, id='foo', body=body)

    @ddt.data(*base.all_users)
    def test_delete_policy(self, user_id):
        volume = self._create_volume(metadata={"foo": "bar"})
        rule_name = policy.DELETE_POLICY
        url = '%s/%s/metadata/foo' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'

        # Relax the GET_POLICY in order to get past that check.
        self.policy.set_rules({policy.GET_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.delete, req,
                                 volume_id=volume.id, id='foo')

    @ddt.data(*base.all_users)
    def test_image_metadata_show_policy(self, user_id):
        image_metadata = {
            "up": "down",
            "left": "right"
        }
        volume = self._create_volume(image_metadata)
        volume = volume.obj_to_primitive()['versioned_object.data']
        rule_name = policy.IMAGE_METADATA_SHOW_POLICY
        url = '%s/%s' % (self.api_path, volume['id'])
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.get_db_volume = mock.MagicMock()
        req.get_db_volume.return_value = volume
        resp_obj = mock.MagicMock(obj={'volume': volume})

        self.assertNotIn('volume_image_metadata', volume.keys())

        self.common_policy_check(user_id, self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.image_controller.show, req,
                                 resp_obj, id=volume['id'], fatal=False)

        if user_id in self.authorized_readers:
            self.assertDictEqual(image_metadata,
                                 volume['volume_image_metadata'])

    @ddt.data(*base.all_users)
    def test_image_metadata_set_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.IMAGE_METADATA_SET_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-set_image_metadata": {
                "metadata": {
                    "image_name": "my_image",
                }
            }
        }

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.image_controller.create, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_image_metadata_remove_policy(self, user_id):
        volume = self._create_volume(image_metadata={"foo": "bar"})
        rule_name = policy.IMAGE_METADATA_REMOVE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-unset_image_metadata": {
                "key": "foo"
            }
        }

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.image_controller.delete, req,
                                 id=volume.id, body=body)

    # NOTE(abishop):
    # The following code is a work in progress, and work is deferred until
    # Yoga. This is because the UPDATE_ADMIN_METADATA_POLICY rule is
    # unchanged in Xena (it's RULE_ADMIN_API). This test will be necessary
    # when RULE_ADMIN_API is deprecated in Yoga.
    #
    # @ddt.data(*base.all_users)
    # def test_update_admin_metadata_policy(self, user_id):
    #     volume = self._create_volume()
    #     rule_name = policy.UPDATE_ADMIN_METADATA_POLICY
    #     url = '%s/%s/action' % (self.api_path, volume.id)
    #     req = fake_api.HTTPRequest.blank(url, version=self.api_version)
    #     req.method = 'POST'
    #     body = {
    #         "os-update_readonly_flag": {
    #             "readonly": True
    #         }
    #     }
    #
    #     # Only this test needs a VolumeActionsController
    #     ext_mgr = extensions.ExtensionManager()
    #     controller = volume_actions.VolumeActionsController(ext_mgr)
    #
    #     # Relax the UPDATE_READONLY_POLICY in order to get past that check.
    #     self.policy.set_rules({va_policy.UPDATE_READONLY_POLICY: ""},
    #                           overwrite=False)
    #
    #     self.common_policy_check(user_id, self.authorized_admins,
    #                              self.unauthorized_admins,
    #                              self.unauthorized_exceptions,
    #                              rule_name,
    #                              controller._volume_readonly_update, req,
    #                              id=volume.id, body=body)


class VolumeMetadataPolicySecureRbacTest(VolumeMetadataPolicyTest):
    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
    ]

    unauthorized_readers = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]

    unauthorized_members = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)


class VolumePolicyTests(test_base.CinderPolicyTests):
    def test_admin_can_get_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(admin_context, path, 'GET')
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v', res_meta['k'])

    def test_owner_can_get_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(user_context, path, 'GET')
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v', res_meta['k'])

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_get_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        response = self._get_request_response(non_owner_context, path, 'GET')
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_create_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k1": "v1"}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_owner_can_create_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k1": "v1"}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_create_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k1": "v1"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_delete_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata/%(key)s' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id,
            'key': 'k'
        }
        response = self._get_request_response(admin_context, path, 'DELETE')
        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_owner_can_delete_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata/%(key)s' % {
            'project_id': user_context.project_id, 'volume_id': volume.id,
            'key': 'k'
        }
        response = self._get_request_response(user_context, path, 'DELETE')
        self.assertEqual(HTTPStatus.OK, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_delete_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata/%(key)s' % {
            'project_id': non_owner_context.project_id,
            'volume_id': volume.id,
            'key': 'k'
        }
        response = self._get_request_response(non_owner_context, path,
                                              'DELETE')
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_update_metadata(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k": "v2"}}
        response = self._get_request_response(admin_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v2', res_meta['k'])

    def test_owner_can_update_metadata(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, metadata={"k": "v"})
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k": "v2"}}
        response = self._get_request_response(user_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)
        res_meta = response.json_body['metadata']
        self.assertIn('k', res_meta)
        self.assertEqual('v2', res_meta['k'])

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_update_metadata_for_others(self, mock_volume):
        owner_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(owner_context, metadata={"k": "v"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/metadata' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"metadata": {"k": "v2"}}
        response = self._get_request_response(non_owner_context, path, 'PUT',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from cinder.policies import base


MANAGE_POLICY = 'snapshot_extension:snapshot_manage'
UNMANAGE_POLICY = 'snapshot_extension:snapshot_unmanage'
LIST_MANAGEABLE_POLICY = 'snapshot_extension:list_manageable'

manageable_snapshots_policies = [
    policy.DocumentedRuleDefault(
        name=LIST_MANAGEABLE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description=
        "List (in detail) of snapshots which are available to manage.",
        operations=[
            {
                'method': 'GET',
                'path': '/manageable_snapshots'
            },
            {
                'method': 'GET',
                'path': '/manageable_snapshots/detail'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=MANAGE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Manage an existing snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/manageable_snapshots'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=UNMANAGE_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="Stop managing a snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots/{snapshot_id}/action (os-unmanage)'
            }
        ]),
]


def list_rules():
    return manageable_snapshots_policies



# Copyright (C) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Handles all requests relating to groups.
"""


from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_utils import uuidutils

from cinder import db
from cinder.db import base
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import fields as c_fields
from cinder.policies import group_actions as gp_action_policy
from cinder.policies import group_snapshot_actions as gsnap_action_policy
from cinder.policies import group_snapshots as gsnap_policy
from cinder.policies import groups as group_policy
from cinder import quota
from cinder import quota_utils
from cinder.scheduler import rpcapi as scheduler_rpcapi
from cinder.volume import api as volume_api
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import volume_types
from cinder.volume import volume_utils


CONF = cfg.CONF

LOG = logging.getLogger(__name__)
GROUP_QUOTAS = quota.GROUP_QUOTAS
VALID_REMOVE_VOL_FROM_GROUP_STATUS = (
    'available',
    'in-use',
    'error',
    'error_deleting')
VALID_ADD_VOL_TO_GROUP_STATUS = (
    'available',
    'in-use')


class API(base.Base):
    """API for interacting with the volume manager for groups."""

    def __init__(self):
        self.scheduler_rpcapi = scheduler_rpcapi.SchedulerAPI()
        self.volume_rpcapi = volume_rpcapi.VolumeAPI()
        self.volume_api = volume_api.API()

        super().__init__()

    def _extract_availability_zone(self, availability_zone):
        raw_zones = self.volume_api.list_availability_zones(enable_cache=True)
        availability_zones = set([az['name'] for az in raw_zones])
        if CONF.storage_availability_zone:
            availability_zones.add(CONF.storage_availability_zone)

        if availability_zone is None:
            if CONF.default_availability_zone:
                availability_zone = CONF.default_availability_zone
            else:
                # For backwards compatibility use the storage_availability_zone
                availability_zone = CONF.storage_availability_zone

        if availability_zone not in availability_zones:
            if CONF.allow_availability_zone_fallback:
                original_az = availability_zone
                availability_zone = (
                    CONF.default_availability_zone or
                    CONF.storage_availability_zone)
                LOG.warning("Availability zone '%(s_az)s' not found, falling "
                            "back to '%(s_fallback_az)s'.",
                            {'s_az': original_az,
                             's_fallback_az': availability_zone})
            else:
                msg = _("Availability zone '%(s_az)s' is invalid.")
                msg = msg % {'s_az': availability_zone}
                raise exception.InvalidInput(reason=msg)

        return availability_zone

    def _update_volumes_host(self, context, group):
        volumes = objects.VolumeList.get_all_by_generic_group(context,
                                                              group.id)
        for vol in volumes:
            # Update the host field for the volume.
            vol.host = group.host
            vol.cluster_name = group.cluster_name
            vol.save()

    def create(self, context, name, description, group_type,
               volume_types, availability_zone=None):
        context.authorize(group_policy.CREATE_POLICY)

        req_volume_types = []
        # NOTE: Admin context is required to get extra_specs of volume_types.
        req_volume_types = (self.db.volume_types_get_by_name_or_id(
            context.elevated(), volume_types))

        if not uuidutils.is_uuid_like(group_type):
            req_group_type = self.db.group_type_get_by_name(context,
                                                            group_type)
        else:
            try:
                req_group_type = self.db.group_type_get(context, group_type)
            except exception.GroupTypeNotFound:
                # check again if we get this group type by uuid-like name
                try:
                    req_group_type = self.db.group_type_get_by_name(
                        context, group_type)
                except exception.GroupTypeNotFoundByName:
                    raise exception.GroupTypeNotFound(group_type_id=group_type)

        availability_zone = self._extract_availability_zone(availability_zone)
        kwargs = {'user_id': context.user_id,
                  'project_id': context.project_id,
                  'availability_zone': availability_zone,
                  'status': c_fields.GroupStatus.CREATING,
                  'name': name,
                  'description': description,
                  'volume_type_ids': [t['id'] for t in req_volume_types],
                  'group_type_id': req_group_type['id'],
                  'replication_status': c_fields.ReplicationStatus.DISABLED}
        try:
            reservations = GROUP_QUOTAS.reserve(context,
                                                project_id=context.project_id,
                                                groups=1)
        except exception.OverQuota as e:
            quota_utils.process_reserve_over_quota(context, e,
                                                   resource='groups')
        group = None
        try:
            group = objects.Group(context=context, **kwargs)
            group.create()
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Error occurred when creating group"
                          " %s.", name)
                GROUP_QUOTAS.rollback(context, reservations)

        request_spec_list = []
        filter_properties_list = []
        for req_volume_type in req_volume_types:
            request_spec = {'volume_type': req_volume_type.copy(),
                            'group_id': group.id}
            filter_properties = {}
            request_spec_list.append(request_spec)
            filter_properties_list.append(filter_properties)

        group_spec = {'group_type': req_group_type.copy(),
                      'group_id': group.id}
        group_filter_properties = {}

        # Update quota for groups
        GROUP_QUOTAS.commit(context, reservations)

        self._cast_create_group(context, group,
                                group_spec,
                                request_spec_list,
                                group_filter_properties,
                                filter_properties_list)

        return group

    def create_from_src(self, context, name, description=None,
                        group_snapshot_id=None, source_group_id=None):
        context.authorize(group_policy.CREATE_POLICY)

        # Populate group_type_id and volume_type_ids
        group_type_id = None
        volume_type_ids = []
        size = 0
        if group_snapshot_id:
            grp_snap = self.get_group_snapshot(context, group_snapshot_id)
            group_type_id = grp_snap.group_type_id
            grp_snap_src_grp = self.get(context, grp_snap.group_id)
            volume_type_ids = [vt.id for vt in grp_snap_src_grp.volume_types]
            snapshots = objects.SnapshotList.get_all_for_group_snapshot(
                context, group_snapshot_id)
            size = sum(s.volume.size for s in snapshots)
        elif source_group_id:
            source_group = self.get(context, source_group_id)
            group_type_id = source_group.group_type_id
            volume_type_ids = [vt.id for vt in source_group.volume_types]
            source_vols = objects.VolumeList.get_all_by_generic_group(
                context, source_group.id)
            size = sum(v.size for v in source_vols)

        kwargs = {
            'user_id': context.user_id,
            'project_id': context.project_id,
            'status': c_fields.GroupStatus.CREATING,
            'name': name,
            'description': description,
            'group_snapshot_id': group_snapshot_id,
            'source_group_id': source_group_id,
            'group_type_id': group_type_id,
            'volume_type_ids': volume_type_ids,
            'replication_status': c_fields.ReplicationStatus.DISABLED
        }
        try:
            reservations = GROUP_QUOTAS.reserve(context,
                                                project_id=context.project_id,
                                                groups=1)
        except exception.OverQuota as e:
            quota_utils.process_reserve_over_quota(context, e,
                                                   resource='groups')
        group = None
        try:
            group = objects.Group(context=context, **kwargs)
            group.create(group_snapshot_id=group_snapshot_id,
                         source_group_id=source_group_id)
        except exception.GroupNotFound:
            with excutils.save_and_reraise_exception():
                LOG.error("Source Group %(source_group)s not found when "
                          "creating group %(group)s from source.",
                          {'group': name, 'source_group': source_group_id})
                GROUP_QUOTAS.rollback(context, reservations)
        except exception.GroupSnapshotNotFound:
            with excutils.save_and_reraise_exception():
                LOG.error("Group snapshot %(group_snap)s not found when "
                          "creating group %(group)s from source.",
                          {'group': name, 'group_snap': group_snapshot_id})
                GROUP_QUOTAS.rollback(context, reservations)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Error occurred when creating group"
                          " %(group)s from group_snapshot %(grp_snap)s.",
                          {'group': name, 'grp_snap': group_snapshot_id})
                GROUP_QUOTAS.rollback(context, reservations)

        # Update quota for groups
        GROUP_QUOTAS.commit(context, reservations)

        # NOTE(tommylikehu): We wrap the size inside of the attribute
        # 'volume_properties' as scheduler's filter logic are all designed
        # based on this attribute.
        kwargs = {'group_id': group.id,
                  'volume_properties': objects.VolumeProperties(size=size)}

        host = group.resource_backend
        if not host or not self.scheduler_rpcapi.validate_host_capacity(
                context, host, objects.RequestSpec(**kwargs)):
            msg = _("No valid host to create group %s.") % group.id
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        group.assert_not_frozen()

        if group_snapshot_id:
            self._create_group_from_group_snapshot(context, group,
                                                   group_snapshot_id)
        elif source_group_id:
            self._create_group_from_source_group(context, group,
                                                 source_group_id)

        return group

    def _create_group_from_group_snapshot(self, context, group,
                                          group_snapshot_id):
        try:
            group_snapshot = objects.GroupSnapshot.get_by_id(
                context, group_snapshot_id)
            snapshots = objects.SnapshotList.get_all_for_group_snapshot(
                context, group_snapshot.id)

            if not snapshots:
                msg = _("Group snapshot is empty. No group will be created.")
                raise exception.InvalidGroup(reason=msg)

            for snapshot in snapshots:
                kwargs = {}
                kwargs['availability_zone'] = group.availability_zone
                kwargs['group_snapshot'] = group_snapshot
                kwargs['group'] = group
                kwargs['snapshot'] = snapshot
                volume_type_id = snapshot.volume_type_id
                if volume_type_id:
                    kwargs['volume_type'] = (
                        objects.VolumeType.get_by_name_or_id(
                            context, volume_type_id))
                    # Create group volume_type mapping entries
                    try:
                        db.group_volume_type_mapping_create(context, group.id,
                                                            volume_type_id)
                    except exception.GroupVolumeTypeMappingExists:
                        # Only need to create one group volume_type mapping
                        # entry for the same combination, skipping.
                        LOG.info("A mapping entry already exists for group"
                                 " %(grp)s and volume type %(vol_type)s. "
                                 "Do not need to create again.",
                                 {'grp': group.id,
                                  'vol_type': volume_type_id})

                # Since group snapshot is passed in, the following call will
                # create a db entry for the volume, but will not call the
                # volume manager to create a real volume in the backend yet.
                # If error happens, taskflow will handle rollback of quota
                # and removal of volume entry in the db.
                try:
                    self.volume_api.create(context,
                                           snapshot.volume_size,
                                           None,
                                           None,
                                           **kwargs)
                except exception.CinderException:
                    with excutils.save_and_reraise_exception():
                        LOG.error("Error occurred when creating volume "
                                  "entry from snapshot in the process of "
                                  "creating group %(group)s "
                                  "from group snapshot %(group_snap)s.",
                                  {'group': group.id,
                                   'group_snap': group_snapshot.id})
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    volumes = objects.VolumeList.get_all_by_generic_group(
                        context, group.id)
                    for vol in volumes:
                        # NOTE(tommylikehu): `delete` is used here in order to
                        # revert consumed quota.
                        self.volume_api.delete(context, vol)
                    group.destroy()
                finally:
                    LOG.error("Error occurred when creating group "
                              "%(group)s from group snapshot %(group_snap)s.",
                              {'group': group.id,
                               'group_snap': group_snapshot.id})

        self._update_volumes_host(context, group)

        self.volume_rpcapi.create_group_from_src(
            context, group, group_snapshot)

    def _create_group_from_source_group(self, context, group,
                                        source_group_id):
        try:
            source_group = objects.Group.get_by_id(context,
                                                   source_group_id)
            source_vols = objects.VolumeList.get_all_by_generic_group(
                context, source_group.id)

            if not source_vols:
                msg = _("Source Group is empty. No group "
                        "will be created.")
                raise exception.InvalidGroup(reason=msg)

            for source_vol in source_vols:
                kwargs = {}
                kwargs['availability_zone'] = group.availability_zone
                kwargs['source_group'] = source_group
                kwargs['group'] = group
                kwargs['source_volume'] = source_vol
                volume_type_id = source_vol.volume_type_id
                if volume_type_id:
                    kwargs['volume_type'] = (
                        objects.VolumeType.get_by_name_or_id(
                            context, volume_type_id))
                    # Create group volume_type mapping entries
                    try:
                        db.group_volume_type_mapping_create(context, group.id,
                                                            volume_type_id)
                    except exception.GroupVolumeTypeMappingExists:
                        # Only need to create one group volume_type mapping
                        # entry for the same combination, skipping.
                        LOG.info("A mapping entry already exists for group"
                                 " %(grp)s and volume type %(vol_type)s. "
                                 "Do not need to create again.",
                                 {'grp': group.id,
                                  'vol_type': volume_type_id})

                # Since source_group is passed in, the following call will
                # create a db entry for the volume, but will not call the
                # volume manager to create a real volume in the backend yet.
                # If error happens, taskflow will handle rollback of quota
                # and removal of volume entry in the db.
                try:
                    self.volume_api.create(context,
                                           source_vol.size,
                                           None,
                                           None,
                                           **kwargs)
                except exception.CinderException:
                    with excutils.save_and_reraise_exception():
                        LOG.error("Error occurred when creating cloned "
                                  "volume in the process of creating "
                                  "group %(group)s from "
                                  "source group %(source_group)s.",
                                  {'group': group.id,
                                   'source_group': source_group.id})
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    volumes = objects.VolumeList.get_all_by_generic_group(
                        context, group.id)
                    for vol in volumes:
                        # NOTE(tommylikehu): `delete` is used here in order to
                        # revert consumed quota.
                        self.volume_api.delete(context, vol)
                    group.destroy()
                finally:
                    LOG.error("Error occurred when creating "
                              "group %(group)s from source group "
                              "%(source_group)s.",
                              {'group': group.id,
                               'source_group': source_group.id})

        self._update_volumes_host(context, group)

        self.volume_rpcapi.create_group_from_src(context, group,
                                                 None, source_group)

    def _cast_create_group(self, context, group,
                           group_spec,
                           request_spec_list,
                           group_filter_properties,
                           filter_properties_list):

        try:
            for request_spec in request_spec_list:
                volume_type = request_spec.get('volume_type')
                volume_type_id = None
                if volume_type:
                    volume_type_id = volume_type.get('id')

                specs = {}
                if volume_type_id:
                    qos_specs = volume_types.get_volume_type_qos_specs(
                        volume_type_id)
                    specs = qos_specs['qos_specs']
                if not specs:
                    # to make sure we don't pass empty dict
                    specs = None

                volume_properties = {
                    'size': 0,  # Need to populate size for the scheduler
                    'user_id': context.user_id,
                    'project_id': context.project_id,
                    'status': 'creating',
                    'attach_status': 'detached',
                    'encryption_key_id': request_spec.get('encryption_key_id'),
                    'display_description': request_spec.get('description'),
                    'display_name': request_spec.get('name'),
                    'volume_type_id': volume_type_id,
                    'group_type_id': group.group_type_id,
                    'availability_zone': group.availability_zone
                }

                request_spec['volume_properties'] = volume_properties
                request_spec['qos_specs'] = specs

            group_properties = {
                'size': 0,  # Need to populate size for the scheduler
                'user_id': context.user_id,
                'project_id': context.project_id,
                'status': 'creating',
                'display_description': group_spec.get('description'),
                'display_name': group_spec.get('name'),
                'group_type_id': group.group_type_id,
            }

            group_spec['volume_properties'] = group_properties
            group_spec['qos_specs'] = None

        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    group.destroy()
                finally:
                    LOG.error("Error occurred when building request spec "
                              "list for group %s.", group.id)

        # Cast to the scheduler and let it handle whatever is needed
        # to select the target host for this group.
        self.scheduler_rpcapi.create_group(
            context,
            group,
            group_spec=group_spec,
            request_spec_list=request_spec_list,
            group_filter_properties=group_filter_properties,
            filter_properties_list=filter_properties_list)

    def update_quota(self, context, group, num, project_id=None):
        reserve_opts = {'groups': num}
        try:
            reservations = GROUP_QUOTAS.reserve(context,
                                                project_id=project_id,
                                                **reserve_opts)
            if reservations:
                GROUP_QUOTAS.commit(context, reservations)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                try:
                    group.destroy()
                    if isinstance(e, exception.OverQuota):
                        quota_utils.process_reserve_over_quota(
                            context, e, resource='groups')
                finally:
                    LOG.error("Failed to update quota for group %s.", group.id)

    def delete(self, context, group, delete_volumes=False):
        context.authorize(gp_action_policy.DELETE_POLICY, target_obj=group)
        if not group.host:
            self.update_quota(context, group, -1, group.project_id)

            LOG.debug("No host for group %s. Deleting from "
                      "the database.", group.id)
            group.destroy()

            return

        group.assert_not_frozen()

        if not delete_volumes and group.status not in (
                [c_fields.GroupStatus.AVAILABLE,
                 c_fields.GroupStatus.ERROR]):
            msg = _("Group status must be available or error, "
                    "but current status is: %s") % group.status
            raise exception.InvalidGroup(reason=msg)

        # NOTE(tommylikehu): Admin context is required to load group snapshots.
        with group.obj_as_admin():
            if group.group_snapshots:
                raise exception.InvalidGroup(
                    reason=_("Group has existing snapshots."))

        # TODO(smcginnis): Add conditional update handling for volumes
        # Should probably utilize the volume_api.delete code to handle
        # cascade snapshot deletion and force delete.
        volumes = self.db.volume_get_all_by_generic_group(context.elevated(),
                                                          group.id)
        if volumes and not delete_volumes:
            msg = (_("Group %s still contains volumes. "
                     "The delete-volumes flag is required to delete it.")
                   % group.id)
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        volumes_model_update = []
        for volume in volumes:
            if volume['attach_status'] == "attached":
                msg = _("Volume in group %s is attached. "
                        "Need to detach first.") % group.id
                LOG.error(msg)
                raise exception.InvalidGroup(reason=msg)

            snapshots = objects.SnapshotList.get_all_for_volume(context,
                                                                volume['id'])
            if snapshots:
                msg = _("Volume in group still has "
                        "dependent snapshots.")
                LOG.error(msg)
                raise exception.InvalidGroup(reason=msg)

            volumes_model_update.append({'id': volume['id'],
                                         'status': 'deleting'})

        self.db.volumes_update(context, volumes_model_update)

        if delete_volumes:
            # We're overloading the term "delete_volumes" somewhat to also
            # mean to delete the group regardless of the state.
            expected = {}
        else:
            expected = {'status': (c_fields.GroupStatus.AVAILABLE,
                                   c_fields.GroupStatus.ERROR)}
        filters = [~db.group_has_group_snapshot_filter(),
                   ~db.group_has_volumes_filter(
                       attached_or_with_snapshots=delete_volumes),
                   ~db.group_creating_from_src(group_id=group.id)]
        values = {'status': c_fields.GroupStatus.DELETING}

        if not group.conditional_update(values, expected, filters):
            if delete_volumes:
                reason = _('Group status must be available or error and must '
                           'not have dependent group snapshots')
            else:
                reason = _('Group must not have attached volumes, volumes '
                           'with snapshots, or dependent group snapshots')
            msg = _('Cannot delete group %(id)s. %(reason)s, and '
                    'it cannot be the source for an ongoing group or group '
                    'snapshot creation.') % {
                'id': group.id, 'reason': reason}
            raise exception.InvalidGroup(reason=msg)

        self.volume_rpcapi.delete_group(context, group)

    def update(self, context, group, name, description,
               add_volumes, remove_volumes):
        """Update group."""
        context.authorize(group_policy.UPDATE_POLICY, target_obj=group)
        # Validate name.
        if name == group.name:
            name = None

        # Validate description.
        if description == group.description:
            description = None

        add_volumes_list = []
        remove_volumes_list = []
        if add_volumes:
            add_volumes = add_volumes.strip(',')
            add_volumes_list = add_volumes.split(',')
        if remove_volumes:
            remove_volumes = remove_volumes.strip(',')
            remove_volumes_list = remove_volumes.split(',')

        invalid_uuids = []
        for uuid in add_volumes_list:
            if uuid in remove_volumes_list:
                invalid_uuids.append(uuid)
        if invalid_uuids:
            msg = _("UUIDs %s are in both add and remove volume "
                    "list.") % invalid_uuids
            raise exception.InvalidVolume(reason=msg)

        volumes = self.db.volume_get_all_by_generic_group(context, group.id)

        # Validate volumes in add_volumes and remove_volumes.
        add_volumes_new = ""
        remove_volumes_new = ""
        if add_volumes_list:
            add_volumes_new = self._validate_add_volumes(
                context, volumes, add_volumes_list, group)
        if remove_volumes_list:
            remove_volumes_new = self._validate_remove_volumes(
                volumes, remove_volumes_list, group)

        if (name is None and description is None and not add_volumes_new and
                not remove_volumes_new):
            msg = (_("Cannot update group %(group_id)s "
                     "because no valid name, description, add_volumes, "
                     "or remove_volumes were provided.") %
                   {'group_id': group.id})
            raise exception.InvalidGroup(reason=msg)

        expected = {}
        fields = {'updated_at': timeutils.utcnow()}

        # Update name and description in db now. No need to
        # to send them over through an RPC call.
        if name is not None:
            fields['name'] = name
        if description is not None:
            fields['description'] = description
        if not add_volumes_new and not remove_volumes_new:
            # Only update name or description. Set status to available.
            fields['status'] = c_fields.GroupStatus.AVAILABLE
        else:
            expected['status'] = c_fields.GroupStatus.AVAILABLE
            fields['status'] = c_fields.GroupStatus.UPDATING

        if not group.conditional_update(fields, expected):
            msg = _("Group status must be available.")
            raise exception.InvalidGroup(reason=msg)

        # Do an RPC call only if the update request includes
        # adding/removing volumes. add_volumes_new and remove_volumes_new
        # are strings of volume UUIDs separated by commas with no spaces
        # in between.
        if add_volumes_new or remove_volumes_new:
            self.volume_rpcapi.update_group(
                context, group,
                add_volumes=add_volumes_new,
                remove_volumes=remove_volumes_new)

    def _validate_remove_volumes(self, volumes, remove_volumes_list, group):
        # Validate volumes in remove_volumes.
        remove_volumes_new = ""
        for volume in volumes:
            if volume['id'] in remove_volumes_list:
                if volume['status'] not in VALID_REMOVE_VOL_FROM_GROUP_STATUS:
                    msg = (_("Cannot remove volume %(volume_id)s from "
                             "group %(group_id)s because volume "
                             "is in an invalid state: %(status)s. Valid "
                             "states are: %(valid)s.") %
                           {'volume_id': volume['id'],
                            'group_id': group.id,
                            'status': volume['status'],
                            'valid': VALID_REMOVE_VOL_FROM_GROUP_STATUS})
                    raise exception.InvalidVolume(reason=msg)
                # Volume currently in group. It will be removed from group.
                if remove_volumes_new:
                    remove_volumes_new += ","
                remove_volumes_new += volume['id']

        for rem_vol in remove_volumes_list:
            if rem_vol not in remove_volumes_new:
                msg = (_("Cannot remove volume %(volume_id)s from "
                         "group %(group_id)s because it "
                         "is not in the group.") %
                       {'volume_id': rem_vol,
                        'group_id': group.id})
                raise exception.InvalidVolume(reason=msg)

        return remove_volumes_new

    def _validate_add_volumes(self, context, volumes, add_volumes_list, group):
        add_volumes_new = ""
        for volume in volumes:
            if volume['id'] in add_volumes_list:
                # Volume already in group. Remove from add_volumes.
                add_volumes_list.remove(volume['id'])

        for add_vol in add_volumes_list:
            try:
                add_vol_ref = objects.Volume.get_by_id(context, add_vol)
            except exception.VolumeNotFound:
                msg = (_("Cannot add volume %(volume_id)s to "
                         "group %(group_id)s because volume cannot be "
                         "found.") %
                       {'volume_id': add_vol,
                        'group_id': group.id})
                raise exception.InvalidVolume(reason=msg)
            orig_group = add_vol_ref.group_id
            if orig_group:
                # If volume to be added is already in the group to be updated,
                # it should have been removed from the add_volumes_list in the
                # beginning of this function. If we are here, it means it is
                # in a different group.
                msg = (_("Cannot add volume %(volume_id)s to group "
                         "%(group_id)s because it is already in "
                         "group %(orig_group)s.") %
                       {'volume_id': add_vol_ref.id,
                        'group_id': group.id,
                        'orig_group': orig_group})
                raise exception.InvalidVolume(reason=msg)
            if add_vol_ref:
                if add_vol_ref.project_id != group.project_id:
                    msg = (_("Cannot add volume %(volume_id)s to group "
                             "%(group_id)s as they belong to different "
                             "projects.") %
                           {'volume_id': add_vol_ref.id,
                            'group_id': group.id})
                    raise exception.InvalidVolume(reason=msg)
                add_vol_type_id = add_vol_ref.volume_type_id
                if not add_vol_type_id:
                    msg = (_("Cannot add volume %(volume_id)s to group "
                             "%(group_id)s because it has no volume "
                             "type.") %
                           {'volume_id': add_vol_ref.id,
                            'group_id': group.id})
                    raise exception.InvalidVolume(reason=msg)
                vol_type_ids = [v_type.id for v_type in group.volume_types]
                if add_vol_type_id not in vol_type_ids:
                    msg = (_("Cannot add volume %(volume_id)s to group "
                             "%(group_id)s because volume type "
                             "%(volume_type)s is not supported by the "
                             "group.") %
                           {'volume_id': add_vol_ref.id,
                            'group_id': group.id,
                            'volume_type': add_vol_type_id})
                    raise exception.InvalidVolume(reason=msg)
                if (add_vol_ref.status not in
                        VALID_ADD_VOL_TO_GROUP_STATUS):
                    msg = (_("Cannot add volume %(volume_id)s to group "
                             "%(group_id)s because volume is in an "
                             "invalid state: %(status)s. Valid states are: "
                             "%(valid)s.") %
                           {'volume_id': add_vol_ref.id,
                            'group_id': group.id,
                            'status': add_vol_ref.status,
                            'valid': VALID_ADD_VOL_TO_GROUP_STATUS})
                    raise exception.InvalidVolume(reason=msg)

                # group.resource_backend and add_vol_ref.resource_backend are
                # in this format like 'host@backend#pool' in a non-HA
                # deployment and will contain cluster_name in
                # A/A HA deployment.
                vol_host = volume_utils.extract_host(
                    add_vol_ref.resource_backend)
                group_host = volume_utils.extract_host(group.resource_backend)
                if group_host != vol_host:
                    raise exception.InvalidVolume(
                        reason=_("Volume is not local to this node."))

                # Volume exists. It will be added to CG.
                if add_volumes_new:
                    add_volumes_new += ","
                add_volumes_new += add_vol_ref.id

            else:
                msg = (_("Cannot add volume %(volume_id)s to group "
                         "%(group_id)s because volume does not exist.") %
                       {'volume_id': add_vol_ref.id,
                        'group_id': group.id})
                raise exception.InvalidVolume(reason=msg)

        return add_volumes_new

    def get(self, context, group_id):
        group = objects.Group.get_by_id(context, group_id)
        context.authorize(group_policy.GET_POLICY, target_obj=group)
        return group

    def get_all(self, context, filters=None, marker=None, limit=None,
                offset=None, sort_keys=None, sort_dirs=None):
        context.authorize(group_policy.GET_ALL_POLICY)
        if filters is None:
            filters = {}

        if filters:
            LOG.debug("Searching by: %s", filters)

        if (context.is_admin and 'all_tenants' in filters):
            del filters['all_tenants']
            groups = objects.GroupList.get_all(
                context, filters=filters, marker=marker, limit=limit,
                offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        else:
            groups = objects.GroupList.get_all_by_project(
                context, context.project_id, filters=filters, marker=marker,
                limit=limit, offset=offset, sort_keys=sort_keys,
                sort_dirs=sort_dirs)
        return groups

    def reset_status(self, context, group, status):
        """Reset status of generic group"""
        context.authorize(gp_action_policy.RESET_STATUS, target_obj=group)
        field = {'updated_at': timeutils.utcnow(),
                 'status': status}
        group.update(field)
        group.save()

    def create_group_snapshot(self, context, group, name, description):
        context.authorize(gsnap_policy.CREATE_POLICY, target_obj=group)
        group.assert_not_frozen()
        options = {'group_id': group.id,
                   'user_id': context.user_id,
                   'project_id': context.project_id,
                   'status': "creating",
                   'name': name,
                   'description': description,
                   'group_type_id': group.group_type_id}

        group_snapshot = None
        group_snapshot_id = None
        try:
            group_snapshot = objects.GroupSnapshot(context, **options)
            group_snapshot.create()
            group_snapshot_id = group_snapshot.id

            snap_name = group_snapshot.name
            snap_desc = group_snapshot.description
            with group.obj_as_admin():
                self.volume_api.create_snapshots_in_db(
                    context, group.volumes, snap_name, snap_desc,
                    None, group_snapshot_id)

        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    # If the group_snapshot has been created
                    if group_snapshot.obj_attr_is_set('id'):
                        group_snapshot.destroy()
                finally:
                    LOG.error("Error occurred when creating group_snapshot"
                              " %s.", group_snapshot_id)

        self.volume_rpcapi.create_group_snapshot(context, group_snapshot)

        return group_snapshot

    def delete_group_snapshot(self, context, group_snapshot, force=False):
        context.authorize(gsnap_policy.DELETE_POLICY,
                          target_obj=group_snapshot)
        group_snapshot.assert_not_frozen()
        values = {'status': 'deleting'}
        expected = {'status': ('available', 'error')}
        filters = [~db.group_creating_from_src(
                   group_snapshot_id=group_snapshot.id)]
        res = group_snapshot.conditional_update(values, expected, filters)

        if not res:
            msg = _('GroupSnapshot status must be available or error, and no '
                    'Group can be currently using it as source for its '
                    'creation.')
            raise exception.InvalidGroupSnapshot(reason=msg)

        snapshots = objects.SnapshotList.get_all_for_group_snapshot(
            context, group_snapshot.id)

        # TODO(xyang): Add a new db API to update all snapshots statuses
        # in one db API call.
        for snap in snapshots:
            snap.status = c_fields.SnapshotStatus.DELETING
            snap.save()

        self.volume_rpcapi.delete_group_snapshot(context.elevated(),
                                                 group_snapshot)

    def update_group_snapshot(self, context, group_snapshot, fields):
        context.authorize(gsnap_policy.UPDATE_POLICY,
                          target_obj=group_snapshot)
        group_snapshot.update(fields)
        group_snapshot.save()

    def get_group_snapshot(self, context, group_snapshot_id):
        group_snapshot = objects.GroupSnapshot.get_by_id(context,
                                                         group_snapshot_id)
        context.authorize(gsnap_policy.GET_POLICY,
                          target_obj=group_snapshot)

        return group_snapshot

    def get_all_group_snapshots(self, context, filters=None, marker=None,
                                limit=None, offset=None, sort_keys=None,
                                sort_dirs=None):
        context.authorize(gsnap_policy.GET_ALL_POLICY)
        filters = filters or {}

        if context.is_admin and 'all_tenants' in filters:
            # Need to remove all_tenants to pass the filtering below.
            del filters['all_tenants']
            group_snapshots = objects.GroupSnapshotList.get_all(
                context, filters=filters, marker=marker, limit=limit,
                offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        else:
            group_snapshots = objects.GroupSnapshotList.get_all_by_project(
                context.elevated(), context.project_id, filters=filters,
                marker=marker, limit=limit, offset=offset, sort_keys=sort_keys,
                sort_dirs=sort_dirs)
        return group_snapshots

    def reset_group_snapshot_status(self, context, gsnapshot, status):
        """Reset status of group snapshot"""

        context.authorize(gsnap_action_policy.RESET_STATUS,
                          target_obj=gsnapshot)
        field = {'updated_at': timeutils.utcnow(),
                 'status': status}
        gsnapshot.update(field)
        gsnapshot.save()

    def _check_type(self, group):
        if not group.is_replicated:
            msg = _("Group %s is not a replication group type.") % group.id
            LOG.error(msg)
            raise exception.InvalidGroupType(reason=msg)

        for vol_type in group.volume_types:
            if not volume_utils.is_replicated_spec(vol_type.extra_specs):
                msg = _("Volume type %s does not have 'replication_enabled' "
                        "spec key set to '<is> True'.") % vol_type.id
                LOG.error(msg)
                raise exception.InvalidVolumeType(reason=msg)

    # Replication group API (Tiramisu)
    def enable_replication(self, context, group):
        context.authorize(gp_action_policy.ENABLE_REP, target_obj=group)
        self._check_type(group)

        valid_status = [c_fields.GroupStatus.AVAILABLE]
        if group.status not in valid_status:
            params = {'valid': valid_status,
                      'current': group.status,
                      'id': group.id}
            msg = _("Group %(id)s status must be %(valid)s, "
                    "but current status is: %(current)s. "
                    "Cannot enable replication.") % params
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        valid_rep_status = [c_fields.ReplicationStatus.DISABLED,
                            c_fields.ReplicationStatus.ENABLED]
        if group.replication_status not in valid_rep_status:
            params = {'valid': valid_rep_status,
                      'current': group.replication_status,
                      'id': group.id}
            msg = _("Group %(id)s replication status must be %(valid)s, "
                    "but current status is: %(current)s. "
                    "Cannot enable replication.") % params
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        volumes = objects.VolumeList.get_all_by_generic_group(
            context.elevated(), group.id)

        valid_status = ['available', 'in-use']
        for vol in volumes:
            if vol.status not in valid_status:
                params = {'valid': valid_status,
                          'current': vol.status,
                          'id': vol.id}
                msg = _("Volume %(id)s status must be %(valid)s, "
                        "but current status is: %(current)s. "
                        "Cannot enable replication.") % params
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)
                # replication_status could be set to enabled when volume is
                # created and the mirror is built.
            if vol.replication_status not in valid_rep_status:
                params = {'valid': valid_rep_status,
                          'current': vol.replication_status,
                          'id': vol.id}
                msg = _("Volume %(id)s replication status must be %(valid)s, "
                        "but current status is: %(current)s. "
                        "Cannot enable replication.") % params
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)

            vol.replication_status = c_fields.ReplicationStatus.ENABLING
            vol.save()

        group.replication_status = c_fields.ReplicationStatus.ENABLING
        group.save()

        self.volume_rpcapi.enable_replication(context, group)

    def disable_replication(self, context, group):
        context.authorize(gp_action_policy.DISABLE_REP, target_obj=group)
        self._check_type(group)

        valid_status = [c_fields.GroupStatus.AVAILABLE,
                        c_fields.GroupStatus.ERROR]
        if group.status not in valid_status:
            params = {'valid': valid_status,
                      'current': group.status,
                      'id': group.id}
            msg = _("Group %(id)s status must be %(valid)s, "
                    "but current status is: %(current)s. "
                    "Cannot disable replication.") % params
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        valid_rep_status = [c_fields.ReplicationStatus.ENABLED,
                            c_fields.ReplicationStatus.ERROR]
        if group.replication_status not in valid_rep_status:
            params = {'valid': valid_rep_status,
                      'current': group.replication_status,
                      'id': group.id}
            msg = _("Group %(id)s replication status must be %(valid)s, "
                    "but current status is: %(current)s. "
                    "Cannot disable replication.") % params
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        volumes = objects.VolumeList.get_all_by_generic_group(
            context.elevated(), group.id)

        for vol in volumes:
            if vol.replication_status not in valid_rep_status:
                params = {'valid': valid_rep_status,
                          'current': vol.replication_status,
                          'id': vol.id}
                msg = _("Volume %(id)s replication status must be %(valid)s, "
                        "but current status is: %(current)s. "
                        "Cannot disable replication.") % params
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)

            vol.replication_status = c_fields.ReplicationStatus.DISABLING
            vol.save()

        group.replication_status = c_fields.ReplicationStatus.DISABLING
        group.save()

        self.volume_rpcapi.disable_replication(context, group)

    def failover_replication(self, context, group,
                             allow_attached_volume=False,
                             secondary_backend_id=None):
        context.authorize(gp_action_policy.FAILOVER_REP, target_obj=group)
        self._check_type(group)

        valid_status = [c_fields.GroupStatus.AVAILABLE]
        if group.status not in valid_status:
            params = {'valid': valid_status,
                      'current': group.status,
                      'id': group.id}
            msg = _("Group %(id)s status must be %(valid)s, "
                    "but current status is: %(current)s. "
                    "Cannot failover replication.") % params
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        valid_rep_status = [c_fields.ReplicationStatus.ENABLED,
                            c_fields.ReplicationStatus.FAILED_OVER]
        if group.replication_status not in valid_rep_status:
            params = {'valid': valid_rep_status,
                      'current': group.replication_status,
                      'id': group.id}
            msg = _("Group %(id)s replication status must be %(valid)s, "
                    "but current status is: %(current)s. "
                    "Cannot failover replication.") % params
            LOG.error(msg)
            raise exception.InvalidGroup(reason=msg)

        volumes = objects.VolumeList.get_all_by_generic_group(
            context.elevated(), group.id)

        valid_status = ['available', 'in-use']
        for vol in volumes:
            if vol.status not in valid_status:
                params = {'valid': valid_status,
                          'current': vol.status,
                          'id': vol.id}
                msg = _("Volume %(id)s status must be %(valid)s, "
                        "but current status is: %(current)s. "
                        "Cannot failover replication.") % params
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)
            if vol.status == 'in-use' and not allow_attached_volume:
                msg = _("Volume %s is attached but allow_attached_volume flag "
                        "is False. Cannot failover replication.") % vol.id
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)
            if vol.replication_status not in valid_rep_status:
                params = {'valid': valid_rep_status,
                          'current': vol.replication_status,
                          'id': vol.id}
                msg = _("Volume %(id)s replication status must be %(valid)s, "
                        "but current status is: %(current)s. "
                        "Cannot failover replication.") % params
                LOG.error(msg)
                raise exception.InvalidVolume(reason=msg)

            vol.replication_status = c_fields.ReplicationStatus.FAILING_OVER
            vol.save()

        group.replication_status = c_fields.ReplicationStatus.FAILING_OVER
        group.save()

        self.volume_rpcapi.failover_replication(context, group,
                                                allow_attached_volume,
                                                secondary_backend_id)

    def list_replication_targets(self, context, group):
        context.authorize(gp_action_policy.LIST_REP, target_obj=group)
        self._check_type(group)

        return self.volume_rpcapi.list_replication_targets(context, group)



#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from unittest import mock
from unittest import TestCase

import requests
from requests import codes

from cinder.exception import NotAuthorized
from cinder.exception import VolumeBackendAPIException
from cinder.tests.unit.fake_constants import VOLUME_NAME
from cinder.tests.unit.volume.drivers.yadro.test_tatlin_common import \
    DummyVolume
from cinder.tests.unit.volume.drivers.yadro.test_tatlin_common import \
    MockResponse
from cinder.volume.drivers.yadro import tatlin_api
from cinder.volume.drivers.yadro.tatlin_client import InitTatlinClient
from cinder.volume.drivers.yadro.tatlin_client import TatlinAccessAPI
from cinder.volume.drivers.yadro.tatlin_client import TatlinClientCommon
from cinder.volume.drivers.yadro.tatlin_client import TatlinClientV23
from cinder.volume.drivers.yadro.tatlin_client import TatlinClientV25
from cinder.volume.drivers.yadro.tatlin_exception import TatlinAPIException


RES_PORTS_RESP = [
    {
        "port": "fc20",
        "port_status": "healthy",
        "port_status_desc": "resource is available",
        "running": [
            "sp-0",
            "sp-1"
        ],
        "wwn": [
            "10:00:14:52:90:00:03:10",
            "10:00:14:52:90:00:03:90"
        ],
        "lun": "scsi-lun-fc20-5",
        "volume": "pty-vol-0d9627cb-c52e-49f1-878c-57c9bc3010c9",
        "lun_index": "5"
    }
]

ALL_HOSTS_RESP = [
    {
        "version": "d6a2d310d9adb16f0d24d5352b5c4837",
        "id": "5e37d335-8fff-4aee-840a-34749301a16a",
        "name": "victoria-fc",
        "port_type": "fc",
        "initiators": [
            "21:00:34:80:0d:6b:aa:e3",
            "21:00:34:80:0d:6b:aa:e2"
        ],
        "tags": [],
        "comment": "",
        "auth": {}
    }
]

RES_MAPPING_RESP = [
    {
        "resource_id": "62bbb941-ba4a-4101-927d-e527ce5ee011",
        "host_id": "5e37d335-8fff-4aee-840a-34749301a16a",
        "mapped_lun_id": 1
    }
]

POOL_LIST_RESPONCE = [
    {
        "id": "7e259486-deb8-4d11-8cb0-e2c5874aaa5e",
        "name": "cinder-pool",
        "status": "ready"
    }
]

VOL_ID = 'cinder-volume-id'

ERROR_VOLUME = [
    {
        "ptyId": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "id": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "name": "cinder-volume-f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "type": "block",
        "poolId": "92c05782-7529-479f-8db7-b9435e1e9a3d",
        "size": 16106127360,
        "maxModifySize": 95330557231104,
        "status": "error",
    }
]

READY_VOLUME = [
    {
        "ptyId": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "id": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "name": "cinder-volume-f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "type": "block",
        "poolId": "92c05782-7529-479f-8db7-b9435e1e9a3d",
        "size": 16106127360,
        "maxModifySize": 95330557231104,
        "status": "ready",
    }
]

RESOURCE_INFORMATION = {
    "ptyId": "62bbb941-ba4a-4101-927d-e527ce5ee011",
    "id": "62bbb941-ba4a-4101-927d-e527ce5ee011",
    "name": "res1",
    "type": "block",
    "poolId": "c46584c5-3113-4cc7-8a72-f9262f32c508",
    "size": 1073741824,
    "maxModifySize": 5761094647808,
    "status": "ready",
    "stat": {
        "used_capacity": 1073741824,
        "mapped_blocks": 0,
        "dedup_count": 0,
        "reduction_ratio": 0
    },
    "lbaFormat": "4kn",
    "volume_id": "pty-vol-62bbb941-ba4a-4101-927d-e527ce5ee011",
    "wwid": "naa.614529011650000c4000800000000004",
    "lun_id": "4",
    "cached": "true",
    "rCacheMode": "enabled",
    "wCacheMode": "enabled",
    "ports": [
        {
            "port": "fc21",
            "port_status": "healthy",
            "port_status_desc":
            "resource is available on all storage controllers",
            "running": [
                "sp-1",
                "sp-0"
            ],
            "wwn": [
                "10:00:14:52:90:00:03:91",
                "10:00:14:52:90:00:03:11"
            ],
            "lun": "scsi-lun-fc21-4",
            "volume": "pty-vol-62bbb941-ba4a-4101-927d-e527ce5ee011",
            "lun_index": "4"
        },
        {
            "port": "fc20",
            "port_status": "healthy",
            "port_status_desc":
            "resource is available on all storage controllers",
            "running": [
                "sp-1",
                "sp-0"
            ],
            "wwn": [
                "10:00:14:52:90:00:03:10",
                "10:00:14:52:90:00:03:90"
            ],
            "lun": "scsi-lun-fc20-4",
            "volume": "pty-vol-62bbb941-ba4a-4101-927d-e527ce5ee011",
            "lun_index": "4"
        }
    ],
    "volume_path": "/dev/mapper/dmc-89382c6c-7cf9-4ff8-bdbb-f438d20c960a",
    "blockSize": "4kn",
    "replication": {
        "is_enabled": False
    }
}

ALL_HOST_GROUP_RESP = [
    {
        "version": "20c28d21549fb7ec5777637f72f50043",
        "id": "314b5546-45da-4c8f-a24c-b615265fbc32",
        "name": "cinder-group",
        "host_ids": [
            "5e37d335-8fff-4aee-840a-34749301a16a"
        ],
        "tags": None,
        "comment": ""
    }
]


class TatlinClientTest(TestCase):
    @mock.patch.object(TatlinAccessAPI, '_authenticate_access')
    def setUp(self, auth_access):
        self.access_api = TatlinAccessAPI('127.0.0.1', 443,
                                          'user', 'passwd', False)
        self.client = TatlinClientV25(self.access_api,
                                      api_retry_count=1,
                                      wait_interval=1,
                                      wait_retry_count=1)

    @mock.patch.object(TatlinAccessAPI, '_authenticate_access')
    @mock.patch.object(TatlinAccessAPI, 'get_tatlin_version')
    def test_different_client_versions(self, version, auth):
        version.side_effect = [(2, 2), (2, 3), (2, 4), (2, 5), (3, 0)]
        args = ['1.2.3.4', 443, 'username', 'password', True, 1, 1, 1]
        self.assertIsInstance(InitTatlinClient(*args), TatlinClientV23)
        self.assertIsInstance(InitTatlinClient(*args), TatlinClientV23)
        self.assertIsInstance(InitTatlinClient(*args), TatlinClientV25)
        self.assertIsInstance(InitTatlinClient(*args), TatlinClientV25)
        self.assertIsInstance(InitTatlinClient(*args), TatlinClientV25)

    @mock.patch.object(requests, 'packages')
    @mock.patch.object(requests, 'session')
    def test_authenticate_success(self, session, packages):
        session().post.return_value = MockResponse({'token': 'ABC'},
                                                   codes.ok)
        TatlinAccessAPI('127.0.0.1', 443, 'user', 'passwd', False)
        session().post.assert_called_once_with(
            'https://127.0.0.1:443/auth/login',
            data={'user': 'user', 'secret': 'passwd'},
            verify=False
        )
        session().headers.update.assert_any_call({'X-Auth-Token': 'ABC'})

        TatlinAccessAPI('127.0.0.1', 443, 'user', 'passwd', True)
        session().headers.update.assert_any_call({'X-Auth-Token': 'ABC'})

    @mock.patch.object(requests, 'session')
    def test_authenticate_fail(self, session):
        session().post.return_value = MockResponse(
            {}, codes.unauthorized)
        self.assertRaises(NotAuthorized,
                          TatlinAccessAPI,
                          '127.0.0.1', 443, 'user', 'passwd', False)

    @mock.patch.object(TatlinAccessAPI, '_authenticate_access')
    @mock.patch.object(requests, 'session')
    def test_send_request(self, session, auth):
        session().request.side_effect = [
            MockResponse({}, codes.ok),
            MockResponse({}, codes.unauthorized),
            MockResponse({}, codes.ok)]

        access_api = TatlinAccessAPI('127.0.0.1', 443, 'user', 'passwd', True)
        access_api.session = session()
        access_api.send_request(tatlin_api.ALL_RESOURCES, {}, 'GET')
        access_api.session.request.assert_called_once_with(
            'GET',
            'https://127.0.0.1:443/' + tatlin_api.ALL_RESOURCES,
            json={},
            verify=True
        )

        access_api.send_request(tatlin_api.ALL_RESOURCES, {}, 'GET')
        self.assertEqual(auth.call_count, 2)
        access_api.session.request.assert_called_with(
            'GET',
            'https://127.0.0.1:443/' + tatlin_api.ALL_RESOURCES,
            json={},
            verify=True
        )

    @mock.patch.object(TatlinAccessAPI, '_authenticate_access')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_tatlin_version(self, send_request, auth):
        send_request.return_value = MockResponse({'build-version': '2.3.0-44'},
                                                 codes.ok)
        access_api = TatlinAccessAPI('127.0.0.1', 443, 'user', 'passwd', True)
        self.assertEqual(access_api.get_tatlin_version(), (2, 3))
        send_request.assert_called_once()

        self.assertEqual(access_api.get_tatlin_version(), (2, 3))
        send_request.assert_called_once()

    @mock.patch.object(TatlinClientCommon, '_is_vol_on_host')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_add_volume_to_host(self,
                                send_request,
                                is_on_host):
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')

        # Success volume already on host
        is_on_host.side_effect = [True]
        self.client.add_vol_to_host(vol.name_id, 10)
        send_request.assert_not_called()

        # Success volume added
        is_on_host.side_effect = [False, True]
        send_request.side_effect = [(MockResponse({}, codes.ok)), ]
        self.client.add_vol_to_host(vol.name_id, 10)

        # Error adding volume to host
        is_on_host.side_effect = [False]
        send_request.side_effect = [
            TatlinAPIException(codes.internal_server_error, ''),
        ]

        with self.assertRaises(TatlinAPIException):
            self.client.add_vol_to_host(vol.name_id, 10)

        # Added successfull but not on host
        is_on_host.side_effect = [False, False]
        send_request.side_effect = [(MockResponse({}, codes.ok)), ]

        with self.assertRaises(VolumeBackendAPIException):
            self.client.add_vol_to_host(vol.name_id, 10)

    @mock.patch.object(TatlinClientCommon, '_is_vol_on_host')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_remove_volume_to_host(self,
                                   send_request,
                                   is_on_host):
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')

        # Success volume not on host
        is_on_host.side_effect = [False]
        self.client.remove_vol_from_host(vol.name_id, 10)
        send_request.assert_not_called()

        # Success volume removed
        is_on_host.side_effect = [True, False]
        send_request.side_effect = [(MockResponse({}, codes.ok)), ]
        self.client.remove_vol_from_host(vol.name_id, 10)

        # Remove from host rise an error
        is_on_host.side_effect = [True, False]
        send_request.side_effect = [
            TatlinAPIException(codes.internal_server_error, ''),
        ]
        with self.assertRaises(TatlinAPIException):
            self.client.remove_vol_from_host(vol.name_id, 10)

        # Removed successfull but still on host
        is_on_host.side_effect = [True, True]
        send_request.side_effect = [(MockResponse({}, codes.ok)), ]

        with self.assertRaises(VolumeBackendAPIException):
            self.client.remove_vol_from_host(vol.name_id, 10)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_is_volume_exist_success(self, send_request):
        send_request.side_effect = [
            (MockResponse(RESOURCE_INFORMATION, codes.ok)),
        ]
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')
        result = self.client.is_volume_exists(vol.name_id)
        self.assertTrue(result)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_is_volume_exist_not_found(self, send_request):
        send_request.return_value = MockResponse(
            RESOURCE_INFORMATION, codes.not_found)
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')
        result = self.client.is_volume_exists(vol.name_id)
        self.assertFalse(result)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_is_volume_exist_unknown_error(self, send_request):
        send_request.return_value = MockResponse(
            {}, codes.internal_server_error)
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')
        with self.assertRaises(VolumeBackendAPIException):
            self.client.is_volume_exists(vol.name_id)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_delete_volume(self, send_request):
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')
        # Success delete
        send_request.side_effect = [(MockResponse({}, codes.ok)), ]
        self.client.delete_volume(vol.name_id)

        # Volume does't exist
        send_request.side_effect = [(MockResponse({}, 404)), ]
        self.client.delete_volume(vol.name_id)

        # Volume delete error
        send_request.side_effect = [
            (MockResponse({}, codes.internal_server_error)),
        ]
        with self.assertRaises(TatlinAPIException):
            self.client.delete_volume(vol.name_id)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_extend_volume(self, send_request):
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')
        # Success delete
        send_request.side_effect = [(MockResponse({}, codes.ok)), ]
        self.client.extend_volume(vol.name_id, 20000)

        # Error
        send_request.side_effect = [
            (MockResponse({}, codes.internal_server_error)),
        ]
        with self.assertRaises(VolumeBackendAPIException):
            self.client.extend_volume(vol.name_id, 20000)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_is_volume_ready(self, send_request):
        send_request.side_effect = [(MockResponse(READY_VOLUME, codes.ok)), ]
        self.assertTrue(self.client.is_volume_ready(VOLUME_NAME))

        send_request.side_effect = [
            (MockResponse(ERROR_VOLUME, codes.ok))
        ]
        self.assertFalse(self.client.is_volume_ready(VOLUME_NAME))

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_host_group_id_success(self, send_request):
        send_request.return_value = MockResponse(
            ALL_HOST_GROUP_RESP, codes.ok)
        self.assertEqual(self.client.get_host_group_id('cinder-group'),
                         '314b5546-45da-4c8f-a24c-b615265fbc32')

    @mock.patch.object(TatlinClientCommon,
                       'is_volume_exists',
                       return_value=True)
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_resource_ports_array(self, send_request, *args):
        send_request.return_value = MockResponse(RES_PORTS_RESP, codes.ok)

        self.assertListEqual(self.client.get_resource_ports_array(VOL_ID),
                             ["fc20"])

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_resource_mapping_negative(self, send_request):
        send_request.return_value = MockResponse(
            {}, codes.internal_server_error)
        self.assertRaises(VolumeBackendAPIException,
                          self.client.get_resource_mapping)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_pool_id_by_name(self, send_request, *args):
        send_request.return_value = MockResponse(POOL_LIST_RESPONCE, codes.ok)
        self.assertEqual(self.client.get_pool_id_by_name('cinder-pool'),
                         '7e259486-deb8-4d11-8cb0-e2c5874aaa5e')

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_all_hosts(self, send_request):
        send_request.return_value = MockResponse({}, codes.ok)
        self.client.get_all_hosts()

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_get_all_hosts_negative(self, send_request):
        send_request.return_value = MockResponse(
            {}, codes.internal_server_error)
        self.assertRaises(VolumeBackendAPIException,
                          self.client.get_all_hosts)



# Copyright (C) 2021 NEC corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Utility module for NEC Driver."""

VERSION = '1.0.0'
CI_WIKI_NAME = 'NEC_V_Cinder_CI'
PARAM_PREFIX = 'nec_v'
VENDOR_NAME = 'NEC'
DRIVER_PREFIX = 'NEC'
DRIVER_FILE_PREFIX = 'nec'
TARGET_PREFIX = 'NEC-'
HDP_VOL_ATTR = 'DP'
HDT_VOL_ATTR = 'DT'
NVOL_LDEV_TYPE = 'DP-VOL'
TARGET_IQN_SUFFIX = '.nec-target'
PAIR_ATTR = 'SS'

DRIVER_INFO = {
    'version': VERSION,
    'proto': '',
    'hba_id': '',
    'hba_id_type': '',
    'msg_id': {
        'target': '',
    },
    'volume_backend_name': '',
    'volume_type': '',
    'param_prefix': PARAM_PREFIX,
    'vendor_name': VENDOR_NAME,
    'driver_prefix': DRIVER_PREFIX,
    'driver_file_prefix': DRIVER_FILE_PREFIX,
    'target_prefix': TARGET_PREFIX,
    'hdp_vol_attr': HDP_VOL_ATTR,
    'hdt_vol_attr': HDT_VOL_ATTR,
    'nvol_ldev_type': NVOL_LDEV_TYPE,
    'target_iqn_suffix': TARGET_IQN_SUFFIX,
    'pair_attr': PAIR_ATTR,
}
